"""
    File: iterative_methods.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: January 6, 2022
    Revised: November 13, 2023
    
    Implementations of voting methods that combine multiple methods
"""

from pref_voting.voting_method import *
from pref_voting.scoring_methods import plurality, borda
from pref_voting.iterative_methods import iterated_removal_cl, instant_runoff, instant_runoff_put, instant_runoff_for_truncated_linear_orders
from pref_voting.profiles import _find_updated_profile, _num_rank

from pref_voting.c1_methods import condorcet, smith_set, copeland, top_cycle
from pref_voting.margin_based_methods import minimax, minimax_scores
from pref_voting.profiles import Profile
from pref_voting.profiles_with_ties import ProfileWithTies
from pref_voting.voting_method_properties import VotingMethodProperties, ElectionTypes

@vm(name = "Daunou",
    input_types = [ElectionTypes.PROFILE])
def daunou(profile, curr_cands=None):
    """Implementation of Daunou's voting method as described in the paper: https://link.springer.com/article/10.1007/s00355-020-01276-w

    If there is a Condorcet winner, then that candidate is the winner.  Otherwise, iteratively remove all Condorcet losers then select the plurality winner from among the remaining candidates.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns:
        A sorted list of candidates

    :Example:

    .. exec_code::

        from pref_voting.profiles import Profile
        from pref_voting.combined_methods import daunou
        from pref_voting.scoring_methods import plurality

        prof = Profile([[1, 3, 2, 0], [0, 2, 3, 1], [1, 3, 0, 2], [3, 1, 0, 2]], [1, 1, 1, 1])

        prof.display()

        daunou.display(prof)
        plurality.display(prof)

    """

    candidates = profile.candidates if curr_cands is None else curr_cands
    cw = profile.condorcet_winner(curr_cands=candidates)
    if cw is not None:
        winners = [cw]
    else:
        cands_survive_it_rem_cl = iterated_removal_cl(profile, curr_cands=curr_cands)
        winners = plurality(profile, curr_cands=cands_survive_it_rem_cl)

    return sorted(winners)


@vm(name = "Blacks",
    input_types = [ElectionTypes.PROFILE])
def blacks(profile, curr_cands=None):
    """If a Condorcet winner exists return that winner. Otherwise, return the Borda winning set.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns:
        A sorted list of candidates

    :Example:

    .. exec_code::

        from pref_voting.profiles import Profile
        from pref_voting.combined_methods import blacks
        from pref_voting.scoring_methods import borda

        prof = Profile([[2, 0, 1], [0, 1, 2], [2, 1, 0], [1, 2, 0]], [1, 1, 1, 1])

        prof.display()

        blacks.display(prof)
        borda.display(prof)


    """

    cw = profile.condorcet_winner(curr_cands=curr_cands)

    if cw is not None:
        winners = [cw]
    else:
        winners = borda(profile, curr_cands=curr_cands)

    return winners

@vm(name = "Smith IRV",
    input_types = [ElectionTypes.PROFILE])
def smith_irv(profile, curr_cands=None):
    """After restricting to the Smith Set, return the Instant Runoff winner.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns:
        A sorted list of candidates

    :Example:

    .. exec_code::

        from pref_voting.profiles import Profile
        from pref_voting.combined_methods import smith_irv
        from pref_voting.iterative_methods import instant_runoff, instant_runoff_put

        prof = Profile([[0, 2, 1, 3], [1, 3, 0, 2], [2, 1, 3, 0], [2, 3, 0, 1]], [1, 1, 1, 1])

        prof.display()

        instant_runoff.display(prof)
        instant_runoff_put.display(prof)
        smith_irv.display(prof)

    """

    smith = smith_set(profile, curr_cands=curr_cands)

    return instant_runoff(profile, curr_cands=smith)

@vm(name = "Smith IRV PUT",
    input_types = [ElectionTypes.PROFILE])
def smith_irv_put(profile, curr_cands=None):
    """After restricting to the Smith Set, return the Instant Runoff winner.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns:
        A sorted list of candidates

    :Example:

    .. exec_code::

        from pref_voting.profiles import Profile
        from pref_voting.combined_methods import smith_irv_put
        from pref_voting.iterative_methods import instant_runoff, instant_runoff_put

        prof = Profile([[0, 2, 1, 3], [1, 3, 0, 2], [2, 1, 3, 0], [2, 3, 0, 1]], [1, 1, 1, 1])

        prof.display()

        instant_runoff.display(prof)
        instant_runoff_put.display(prof)
        smith_irv_put.display(prof)

    """

    smith = smith_set(profile, curr_cands=curr_cands)

    return instant_runoff_put(profile, curr_cands=smith)

@vm(name = "Condorcet IRV",
    input_types = [ElectionTypes.PROFILE])
def condorcet_irv(profile, curr_cands=None, algorithm = "basic", **kwargs):
    """If a Condorcet winner exists, elect that candidate, otherwise return the instant runoff winners.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``
        algorithm (str, optional): The algorithm to use.  Options are "basic" and "recursive".  The default is "basic".

    Returns:
        A sorted list of candidates

    :Example:

    .. exec_code::

        from pref_voting.profiles import Profile
        from pref_voting.combined_methods import condorcet_irv
        from pref_voting.iterative_methods import instant_runoff, instant_runoff_put

        prof = Profile([[0, 2, 1, 3], [1, 3, 0, 2], [2, 1, 3, 0], [2, 3, 0, 1]], [1, 1, 1, 1])

        prof.display()

        instant_runoff.display(prof)
        instant_runoff_put.display(prof)
        condorcet_irv.display(prof)

    """

    cw = profile.condorcet_winner(curr_cands=curr_cands)
    if cw is not None: 
        return [cw]
    else:

        return instant_runoff(profile, curr_cands=curr_cands, algorithm=algorithm,  **kwargs)

@vm(name = "Condorcet IRV PUT",
    input_types = [ElectionTypes.PROFILE])
def condorcet_irv_put(profile, curr_cands=None):
    """If a Condorcet winner exists, elect that candidate, otherwise return the instant runoff put winners.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns:
        A sorted list of candidates

    :Example:

    .. exec_code::

        from pref_voting.profiles import Profile
        from pref_voting.combined_methods import condorcet_irv_put
        from pref_voting.iterative_methods import instant_runoff, instant_runoff_put

        prof = Profile([[0, 2, 1, 3], [1, 3, 0, 2], [2, 1, 3, 0], [2, 3, 0, 1]], [1, 1, 1, 1])

        prof.display()

        instant_runoff.display(prof)
        instant_runoff_put.display(prof)
        condorcet_irv_put.display(prof)

    """

    cw = profile.condorcet_winner(curr_cands=curr_cands)
    if cw is not None: 
        return [cw]
    else:
        return instant_runoff_put(profile, curr_cands=curr_cands)

def compose(vm1, vm2):
    """After restricting the profile to the set of vm1 winners, run vm2

    Args:
        vm1, vm2 (VotingMethod): The voting methods to be composed.

    Returns:
        A VotingMethod that composes vm1 and vm2.

    :Example:

    .. exec_code::

        from pref_voting.profiles import Profile
        from pref_voting.combined_methods import compose
        from pref_voting.scoring_methods import borda
        from pref_voting.c1_methods import copeland

        prof = Profile([[1, 3, 0, 2], [2, 1, 3, 0], [3, 0, 2, 1]], [1, 2, 1])

        prof.display()

        copeland_borda = compose(copeland, borda)

        copeland.display(prof)
        borda.display(prof)
        copeland_borda.display(prof)

    """

    def _vm(edata, curr_cands=None):

        vm1_ws = vm1(edata, curr_cands=curr_cands)

        return vm2(edata, curr_cands=vm1_ws)

    return VotingMethod(_vm, name=f"{vm1.name}-{vm2.name}")


def _compose(vm1, vm2):
    """
    Same as compose, but used to make it easier to document composed voting methods.
    """

    def _vm(edata, curr_cands=None):

        vm1_ws = vm1(edata, curr_cands=curr_cands)

        return vm2(edata, curr_cands=vm1_ws)

    return _vm

@vm(name = "Condorcet Plurality",
    input_types = [ElectionTypes.PROFILE])
def condorcet_plurality(profile, curr_cands = None):
    """Return the Condorcet winner if one exists, otherwise return the plurality winners.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns:
        A sorted list of candidates

    """

    return _compose(condorcet, plurality)(profile, curr_cands=curr_cands)


@vm(name="Smith-Minimax",
    input_types=[ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MARGIN_GRAPH])
def smith_minimax(edata, curr_cands = None):
    """Return the Minimax winner after restricting to the Smith set.

    Args:
        profile (Profile, ProfileWithTies, MarginGraph): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns:
        A sorted list of candidates

    """

    return _compose(top_cycle, minimax)(edata, curr_cands=curr_cands)

@vm(name="Copeland-Local-Borda",
    input_types=[ElectionTypes.PROFILE,  ElectionTypes.MARGIN_GRAPH])
def copeland_local_borda(edata, curr_cands = None):
    """Return the Borda winner after restricting to the Copeland winners.

    Args:
        profile (Profile, MarginGraph): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns:
        A sorted list of candidates

    """
    return _compose(copeland, borda)(edata, curr_cands=curr_cands)

@vm(name="Copeland-Local-Minimax",
    input_types=[ElectionTypes.PROFILE,  ElectionTypes.MARGIN_GRAPH])
def copeland_local_minimax(edata, curr_cands = None):
    """Return the Minimax winner after restricting to the Copeland winners.

    Args:
        profile (Profile, MarginGraph): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns:
        A sorted list of candidates

    """
    return _compose(copeland, minimax)(edata, curr_cands=curr_cands)

def voting_method_with_scoring_tiebreaker(vm, score, name):

    def _vm(profile, curr_cands=None):

        vm_ws = vm(profile, curr_cands=curr_cands)

        if len(vm_ws) == 1: 
            return vm_ws
        
        # get (restricted) rankings
        _rankings, rcounts = profile.rankings_counts

        cands_to_ignore = np.array([c for c in profile.candidates if c not in curr_cands]) if curr_cands is not None else np.array([])

        rankings = _rankings if curr_cands is None else _find_updated_profile(np.array(_rankings), cands_to_ignore, len(profile.candidates))
        
        curr_cands = profile.candidates if curr_cands is None else curr_cands

        # find the candidate scores using the score function
        cand_scores = {c: sum(_num_rank(rankings, rcounts, c, level) * score(len(curr_cands), level) for level in range(1, len(curr_cands) + 1)) for c in curr_cands}
    
        max_ws_score = max([cand_scores[w] for w in vm_ws])

        return sorted([w for w in vm_ws if cand_scores[w] == max_ws_score])

    return _vm 

@vm(name="Copeland-Global-Borda",
    input_types=[ElectionTypes.PROFILE])
def copeland_global_borda(profile, curr_cands=None):
    """From the Copeland winners, return the candidate with the largest *global* Borda score.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns:
        A sorted list of candidates

    """

    return voting_method_with_scoring_tiebreaker(copeland, lambda num_cands, rank : num_cands - rank, "Copeland-Global-Borda")(profile, curr_cands=curr_cands)


@vm(name="Copeland-Global-Minimax",
    input_types=[ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES, ElectionTypes.MARGIN_GRAPH])
def copeland_global_minimax(edata, curr_cands=None):
    """From the Copeland winners, return the candidates with the best *global* Minimax score.

    Args:
        edata (Profile, ProfileWithTies, MarginGraph): Any edata with a Margin method.
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns:
        A sorted list of candidates

    """

    curr_cands = edata.candidates if curr_cands is None else curr_cands

    copeland_ws = copeland(edata, curr_cands=curr_cands)

    mm_scores = minimax_scores(edata, curr_cands=curr_cands)

    best_score = max([mm_scores[c] for c in copeland_ws])

    return sorted([c for c in copeland_ws if mm_scores[c] == best_score])

def faceoff(vm1, vm2):
    """If the vm1 and vm2 winners are the same, return that set of winners. Otherwise, for each choice of a vm1 winner A and vm2 winner B, add to the ultimate winners whichever of A or B is majority preferred to the other (or both if they are tied).

    Args:
        vm1, vm2 (VotingMethod): The voting methods to faceoff.

    Returns:
        A VotingMethod that runs the faceoff of vm1 and vm2.

    """

    def _vm(edata, curr_cands=None):

        curr_cands = edata.candidates if curr_cands is None else curr_cands

        vm1_winners = vm1(edata, curr_cands)
        vm2_winners = vm2(edata, curr_cands)

        if vm1_winners == vm2_winners:
            return vm1_winners
        
        else:
            winners = list()

            for a in vm1_winners:
                for b in vm2_winners:
                    if edata.margin(a,b) > 0:
                        winners.append(a)
                    elif edata.margin(b,a) > 0:
                        winners.append(b)
                    elif edata.margin(a,b) == 0:
                        winners.append(a)
                        winners.append(b) 

            return list(set(winners))

    return VotingMethod(_vm, name=f"{vm1.name}-{vm2.name} Faceoff")

def _faceoff(vm1, vm2):
    """
    Same as faceoff, but used to make it easier to document faceoff voting methods.
    """

    def _vm(edata, curr_cands=None):

        curr_cands = edata.candidates if curr_cands is None else curr_cands

        vm1_winners = vm1(edata, curr_cands)
        vm2_winners = vm2(edata, curr_cands)

        if vm1_winners == vm2_winners:
            return vm1_winners
        
        else:
            winners = list()

            for a in vm1_winners:
                for b in vm2_winners:
                    if edata.margin(a,b) > 0:
                        winners.append(a)
                    elif edata.margin(b,a) > 0:
                        winners.append(b)
                    elif edata.margin(a,b) == 0:
                        winners.append(a)
                        winners.append(b) 

            return list(set(winners))

    return _vm

@vm(name="Borda-Minimax Faceoff",
    input_types=[ElectionTypes.PROFILE])
def borda_minimax_faceoff(edata, curr_cands=None):
    """If the Borda and Minimax winners are the same, return that set of winners. Otherwise, for each choice of a Borda winner A and Minimax winner B, add to the ultimate winners whichever of A or B is majority preferred to the other (or both if they are tied).

    Args:
        profile (Profile, MarginGraph): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns:
        A sorted list of candidates

    ..note:
        Proposed by Edward B. Foley.

    """

    return _faceoff(borda, minimax)(edata, curr_cands=curr_cands)


'''
    File: pairwise_profiles.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: June 3, 2024
    
    Functions to reason about profiles of pairwise comparisons.
'''


from math import ceil
import numpy as np
from numba import jit  
import networkx as nx
from tabulate import tabulate
import matplotlib.pyplot as plt
from pref_voting.weighted_majority_graphs import MajorityGraph, MarginGraph, SupportGraph
from pref_voting.rankings import Ranking
import os

# turn off future warnings.
# getting the following warning when calling tabulate to display a profile: 
# /Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/tabulate.py:1027: FutureWarning: elementwise comparison failed; returning scalar instead, but in the future will perform elementwise comparison
#  if headers == "keys" and not rows:
# see https://stackoverflow.com/questions/40659212/futurewarning-elementwise-comparison-failed-returning-scalar-but-in-the-futur
#
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

class PairwiseBallot:
    
    def __init__(self, comparisons, candidates=None, cmap=None):
        """Constructor method for PairwiseBallot.

        Args:
            comparisons (list): List of tuples, lists, or sets representing pairwise comparisons.
            candidates (list or set, optional): Initial set of candidates. Defaults to None.
            cmap (dict, optional): Mapping of candidates to their names. Defaults to None.
        """
        self._comparisons = []
        
        for comp in comparisons:
            if not isinstance(comp, (tuple, list)) or len(comp) != 2:
                raise ValueError("Each element of the list of comparisons should be a tuple or list of length 2.")
            
            if all(isinstance(comp[i], (int, str)) for i in [0, 1]):
                self._comparisons.append(({comp[0], comp[1]}, {comp[0]}))
            elif all(isinstance(comp[i], (set, list, tuple)) for i in [0, 1]):
                self._comparisons.append((set(comp[0]), set(comp[1])))
            else:
                raise ValueError("Each element of the list of comparisons should be a tuple of sets or lists of candidates.")

        if not self._well_formed_comparisons():
            raise ValueError("The pairwise comparisons are not coherent.")
        
        self.candidates = sorted(list(set(c for menu, _ in self._comparisons for c in menu)) if candidates is None else candidates)
        self.cmap = cmap if cmap is not None else {c: str(c) for c in self.candidates}

    def _well_formed_comparisons(self):
        """Check if the pairwise comparisons are all well-formed.
        
        Returns:
            bool: True if the pairwise comparisons are well-formed, False otherwise.
        """
        for menu, choice in self._comparisons:
            if not choice.issubset(menu):
                return False
        menus = [menu for menu, _ in self._comparisons]
        return len(menus) == len(set(frozenset(menu) for menu in menus))

    def num_comparisons(self): 
        """Return the number of pairwise comparisons"""
        return len(self._comparisons)
    
    def weak_pref(self, c1, c2):
        """Return the revealed weak preference of a menu of choices.

        Args:
            c1 (str or int): First candidate.
            c2 (str or int): Second candidate.

        Returns:
            bool: True if there is a weak preference for c1 over c2, False otherwise.
        """
        return any(c1 in menu and c2 in menu and c1 in choice for menu, choice in self._comparisons)

    def strict_pref(self, c1, c2):
        """Return the revealed strict preference of a menu of choices.

        Args:
            c1 (str or int): First candidate.
            c2 (str or int): Second candidate.

        Returns:
            bool: True if there is a strict preference for c1 over c2, False otherwise.
        """
        return self.weak_pref(c1, c2) and not self.weak_pref(c2, c1)

    def indiff(self, c1, c2):
        """Return the revealed indifference of a menu of choices.

        Args:
            c1 (str or int): First candidate.
            c2 (str or int): Second candidate.

        Returns:
            bool: True if there is indifference between c1 and c2, False otherwise.
        """
        return self.weak_pref(c1, c2) and self.weak_pref(c2, c1)

    def has_comparison(self, c1, c2): 
        """Check if there is a comparison between two candidates.

        Args:
            c1 (str or int): First candidate.
            c2 (str or int): Second candidate.

        Returns:
            bool: True if there is a comparison between c1 and c2, False otherwise.
        """
        return any(c1 in menu and c2 in menu for menu, _ in self._comparisons)
    
    def get_comparison(self, c1, c2): 
        """Get the comparison between two candidates.

        Args:
            c1 (str or int): First candidate.
            c2 (str or int): Second candidate.

        Returns:
            tuple: The comparison between c1 and c2.
        """
        comp = [(menu, choice) for menu, choice in self._comparisons if c1 in menu and c2 in menu]
        return comp[0] if len(comp) == 1 else None
    
    def add_comparison(self, menu, choice):
        """Add a new comparison to the existing comparisons.

        Args:
            menu (set): A set of candidates representing the menu.
            choice (set): A set of candidates representing the choice set.

        Raises:
            ValueError: If the new comparison is not coherent with the existing comparisons.
        """
        new_comparison = (set(menu), set(choice))
        self._comparisons.append(new_comparison)
        if not self._well_formed_comparisons():
            self._comparisons.pop()
            raise ValueError("The new comparison is not well-formed given the existing comparisons.")
        self.candidates = sorted(list(set(c for menu, _ in self._comparisons for c in menu)))

    def add_strict_preference(self, c1, c2):
        """Add a new comparison to the existing comparisons where c1 is strictly preferred to c2.

        Args:
            c1 (int, str): A candidate
            c2 (int, str): A candidate.

        Raises:
            ValueError: If the new comparison is not coherent with the existing comparisons.
        """
        self.add_comparison({c1, c2}, {c1})

    def is_transitive(self, cands): 
        """Return True of the comparisons is transitive on the set cands of candidates"""

        for c1 in cands: 
            for c2 in cands: 
                for c3 in cands: 
                    if self.weak_pref(c1, c2) and self.weak_pref(c2, c3) and not self.weak_pref(c1, c3):
                        #print(f"preference {c1} >= {c2} and {c2} >= {c3} but not {c1} >= {c3}")
                        return False
        return True
    
    def is_quasi_transitive(self, cands): 
        """Return True of the comparisons is transitive on the set cands of candidates"""

        for c1 in cands: 
            for c2 in cands: 
                for c3 in cands: 
                    if self.strict_pref(c1, c2) and self.strict_pref(c2, c3) and not self.strict_pref(c1, c3): 
                        #print(f"Strict preference {c1} > {c2} and {c2} > {c3} but not {c1} > {c3}")
                        return False
        return True

    def to_graph(self, curr_cands=None):
        """Return the majority graph of the pairwise comparisons restricted to the candidates in curr_cands."""
        if curr_cands is None:
            curr_cands = self.candidates
        edges = []
        for c1 in curr_cands:
            for c2 in curr_cands:
                if c1 == c2:
                    continue
                if self.has_comparison(c1, c2) and self.strict_pref(c1, c2):
                    edges.append((c1, c2))
                
        return nx.DiGraph(edges)

    def has_tie(self): 
        """Returns True if there is a tie in the pairwise comparisons."""
        for c1 in self.candidates:
            for c2 in self.candidates:
                if c1 != c2 and self.indiff(c1, c2):
                    return True
        return False
    
    def is_coherent(self):
        """Return True if the comparisons are coherent: If a candidate is compared to another candidate, then that candidate must be compared to all canidates"""

        for c in self.candidates: 
            for menu, _ in self._comparisons: 
                if c in menu: 
                    for c1 in self.candidates: 
                        if c != c1 and not self.has_comparison(c, c1): 
                            return False
        return True


    def cycles(self, curr_cands = None):
        """Returns the cycles in the pairwise comparisons.

        This uses the networkx method ``networkx.find_cycle`` to find the cycles in ``self.mg``.

        """

        comparison_graph = self.to_graph(curr_cands)
        
        return list(nx.simple_cycles(comparison_graph))


    def has_cycle(self, curr_cands = None):
        """Returns True if there is a cycle in the comparison  graph."""

        return len(self.cycles(curr_cands=curr_cands)) != 0

    def to_ranking(self): 
        """Return the comparison as a ranking (return an error if comparisons are not transitive)"""

        assert self.is_transitive(self.candidates), "The comparisons must be transitive to convert to a ranking"
        assert self.is_coherent(), "The comparisons must be coherent to convert to a ranking"

        c1, c2 = self.candidates[0], self.candidates[1]
        ranking = {}
        if self.strict_pref(c1, c2): 
            ranking[c1] = 1
            ranking[c2] = 2
        elif self.strict_pref(c2, c1):
            ranking[c2] = 1
            ranking[c1] = 2
        elif self.indiff(c1, c2): 
            ranking[c2] = 1
            ranking[c1] = 1

        for c in self.candidates: 
            prev_rank = 0
            if c not in ranking.keys():
                ranked_last = True 
                for c2, r in sorted(ranking.items(), key=lambda r: r[1]): 
                    if self.strict_pref(c, c2): 
                        ranking[c] = (prev_rank + r) / 2
                        ranked_last = False
                        break
                    elif self.strict_pref(c2, c): 
                        prev_rank = r
                    elif self.indiff(c, c2): 
                        ranking[c] = r
                        ranked_last = False
                        break
                if ranked_last: 
                    ranking[c] = prev_rank + 1

        r = Ranking(ranking)
        r.normalize_ranks()
        return r
    
    def display(self):
        """Display the pairwise comparisons in a readable format."""
        for menu, choice in self._comparisons:
            menu_str = ", ".join(sorted([self.cmap[c] for c in menu]))
            choice_str = ", ".join(sorted([self.cmap[c] for c in choice]))
            print(f"{{{menu_str}}} -> {{{choice_str}}}")


    def __str__(self):
        """Return the comparisons as a string."""

        str_comparisons = ''
        for menu, choice in self._comparisons:
            menu_str = ", ".join(sorted([self.cmap[c] for c in menu]))
            choice_str = ", ".join(sorted([self.cmap[c] for c in choice]))
            str_comparisons += f"{{{menu_str}}} -> {{{choice_str}}}, "
        return str_comparisons[:-2]
    
    
class PairwiseProfile:
    r"""An anonymous profile of pairwise comparisons.   

    Arguments: 
        pairwise_comparisons: List of comparisons or PairwiseBallot instances.
    """

    def __init__(self, pairwise_comparisons, candidates=None, rcounts=None, cmap=None):
        """Constructor method for PairwiseProfile.

        Args:
            pairwise_comparisons (list): List of lists of pairwise comparisons,  or list of PairwiseBallot instances.
            candidates (list or set, optional): List of candidates. Defaults to None.
            rcounts (list, optional): List of counts for each comparison. Defaults to None.
            cmap (dict, optional): Mapping of candidates to their names. Defaults to None.
        """
        self._pairwise_comparisons = []
        
        for comps in pairwise_comparisons:
            if isinstance(comps, PairwiseBallot):
                self._pairwise_comparisons.append(comps)
            else:
                self._pairwise_comparisons.append(PairwiseBallot(comps, candidates=candidates))
        
        if candidates is None:
            candidates = {c for pc in self._pairwise_comparisons for c in pc.candidates}
        self.candidates = sorted(list(candidates))

        self.cand_to_cidx = {c: idx for idx, c in enumerate(self.candidates)}
        self.cidx_to_cand = {idx: c for c, idx in self.cand_to_cidx.items()}
        
        self._rcounts = rcounts if rcounts is not None else [1] * len(pairwise_comparisons)

        self._tally = np.array([[np.sum([count for pc, count in zip(self._pairwise_comparisons, self._rcounts) if pc.strict_pref(c1, c2)]) for c2 in self.candidates] for c1 in self.candidates])
        
        self.cmap = cmap if cmap is not None else {c: str(c) for c in self.candidates}
                
        self.num_voters = np.sum(self._rcounts)
        """The number of voters in the election."""

    @property
    def comparisons_counts(self):
        """Returns the submitted rankings and the list of counts."""
        return self._pairwise_comparisons, self._rcounts

    @property
    def pairwise_comparisons(self): 
        """Returns a list of all pairwise comparisons"""

        return [comp for compidx,comp in enumerate(self._pairwise_comparisons) 
                for _ in range(self._rcounts[compidx])]

    def support(self, c1, c2):
        """The number of voters that rank `c1` above `c2`.
        
        Args:
            c1 (str or int): The first candidate.
            c2 (str or int): The second candidate.
        
        Returns:
            int: Number of voters that rank `c1` above `c2`.
        """
        return self._tally[self.cand_to_cidx[c1]][self.cand_to_cidx[c2]]
    
    def margin(self, c1, c2):
        """The number of voters that rank `c1` above `c2` minus the number of voters that rank `c2` above `c1`.
        
        Args:
            c1 (str or int): The first candidate.
            c2 (str or int): The second candidate.
        
        Returns:
            int: Margin of votes.
        """
        idx1, idx2 = self.cand_to_cidx[c1], self.cand_to_cidx[c2]
        return self._tally[idx1][idx2] - self._tally[idx2][idx1]
        
    def majority_prefers(self, c1, c2): 
        """Returns true if more voters rank `c1` over `c2` than `c2` over `c1`.
        
        Args:
            c1 (str or int): The first candidate.
            c2 (str or int): The second candidate.
        
        Returns:
            bool: True if `c1` is majority preferred over `c2`, False otherwise.
        """
        return self.margin(c1, c2) > 0

    def is_tied(self, c1, c2):
        """Returns True if `c1` is tied with `c2`. 
        
        Args:
            c1 (str or int): The first candidate.
            c2 (str or int): The second candidate.
        
        Returns:
            bool: True if `c1` is tied with `c2`, False otherwise.
        """
        return self.margin(c1, c2) == 0

    def dominators(self, cand, curr_cands=None): 
        """Returns the list of candidates that are majority preferred to `cand` in the profile restricted to the candidates in `curr_cands`.
        
        Args:
            cand (str or int): The candidate.
            curr_cands (list, optional): List of candidates to consider. Defaults to None.
        
        Returns:
            list: List of candidates that are majority preferred to `cand`.
        """        
        candidates = self.candidates if curr_cands is None else curr_cands
        return [c for c in candidates if self.majority_prefers(c, cand)]

    def dominates(self, cand, curr_cands=None): 
        """Returns the list of candidates that `cand` is majority preferred to in the profile restricted to `curr_cands`.
        
        Args:
            cand (str or int): The candidate.
            curr_cands (list, optional): List of candidates to consider. Defaults to None.
        
        Returns:
            list: List of candidates that `cand` is majority preferred to.
        """
        candidates = self.candidates if curr_cands is None else curr_cands
        return [c for c in candidates if self.majority_prefers(cand, c)]
    
    def copeland_scores(self, curr_cands=None, scores=(1, 0, -1)):
        """The Copeland scores in the profile restricted to the candidates in `curr_cands`.
        
        Args:
            curr_cands (list, optional): List of candidates to consider. Defaults to None.
            scores (tuple, optional): Scores for win, tie, and loss. Defaults to (1, 0, -1).
        
        Returns:
            dict: Dictionary associating each candidate in `curr_cands` with its Copeland score.
        """        
        wscore, tscore, lscore = scores
        candidates = self.candidates if curr_cands is None else curr_cands
        c_scores = {c: 0.0 for c in candidates}
        for c1 in candidates:
            for c2 in candidates:
                if self.majority_prefers(c1, c2):
                    c_scores[c1] += wscore
                elif self.majority_prefers(c2, c1):
                    c_scores[c1] += lscore
                elif c1 != c2:
                    c_scores[c1] += tscore
        return c_scores

    def condorcet_winner(self, curr_cands=None):
        """Returns the Condorcet winner in the profile restricted to `curr_cands` if one exists, otherwise return None.
        
        Args:
            curr_cands (list, optional): List of candidates to consider. Defaults to None.
        
        Returns:
            str or int: Condorcet winner if one exists, otherwise None.
        """
        curr_cands = curr_cands if curr_cands is not None else self.candidates
        for c1 in curr_cands: 
            if all(self.majority_prefers(c1, c2) for c2 in curr_cands if c1 != c2): 
                return c1
        return None

    def weak_condorcet_winner(self, curr_cands=None):
        """Returns a list of the weak Condorcet winners in the profile restricted to `curr_cands`.
        
        Args:
            curr_cands (list, optional): List of candidates to consider. Defaults to None.
        
        Returns:
            list: List of weak Condorcet winners.
        """
        curr_cands = curr_cands if curr_cands is not None else self.candidates
        return [c1 for c1 in curr_cands if not any(self.majority_prefers(c2, c1) for c2 in curr_cands if c1 != c2)]

    def condorcet_loser(self, curr_cands=None):
        """Returns the Condorcet loser in the profile restricted to `curr_cands` if one exists, otherwise return None.
        
        Args:
            curr_cands (list, optional): List of candidates to consider. Defaults to None.
        
        Returns:
            str or int: Condorcet loser if one exists, otherwise None.
        """
        curr_cands = curr_cands if curr_cands is not None else self.candidates
        for c1 in curr_cands: 
            if all(self.majority_prefers(c2, c1) for c2 in curr_cands if c1 != c2): 
                return c1
        return None
    
    def strict_maj_size(self):
        """Returns the strict majority of the number of voters.
        
        Returns:
            int: Size of the strict majority.
        """
        return int(self.num_voters / 2 + 1) if self.num_voters % 2 == 0 else int(ceil(float(self.num_voters) / 2))

    def margin_graph(self): 
        """Returns the margin graph of the profile.
        
        Returns:
            dict: Margin graph of the profile.
        """
        
        return MarginGraph(self.candidates, 
                           [(c1, c2, self.margin(c1, c2)) 
                            for c1 in self.candidates 
                            for c2 in self.candidates 
                            if self.majority_prefers(c1, c2)])

    def majority_graph(self): 
        """Returns the margin graph of the profile.
        
        Returns:
            dict: Margin graph of the profile.
        """
        
        return MajorityGraph(self.candidates, 
                             [(c1, c2) 
                              for c1 in self.candidates 
                              for c2 in self.candidates 
                              if self.majority_prefers(c1, c2)])

    def display(self, cmap=None, style="pretty", curr_cands=None):
        """Display the profile (restricted to `curr_cands`) as an ASCII table.
        
        Args:
            cmap (dict, optional): Mapping of candidates to their names. Defaults to None.
            style (str, optional): Style of the display. Defaults to "pretty".
            curr_cands (list, optional): List of candidates to consider. Defaults to None.
        """
        cmap = cmap if cmap is not None else self.cmap
        comparisons, counts = self.comparisons_counts
        for comp_idx, comps in enumerate(comparisons): 
            print(f'{counts[comp_idx]}: {comps}')

    def __add__(self, other_prof): 
        """Returns the sum of two profiles.
        
        Args:
            other_prof (PairwiseProfile): Another PairwiseProfile instance.
        
        Returns:
            PairwiseProfile: The combined profile.
        """
        assert self.candidates == other_prof.candidates, "The two profiles must have the same candidates"
        combined_comparisons = self._pairwise_comparisons + other_prof._pairwise_comparisons
        combined_rcounts = self._rcounts + other_prof._rcounts
        return PairwiseProfile(combined_comparisons, rcounts=combined_rcounts, candidates=self.candidates)



"""
    File: monotonicity_axioms.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: November 4, 2023
    
    Monotonicity axioms 
"""

from pref_voting.axiom import Axiom
from pref_voting.axiom_helpers import *
from pref_voting.rankings import Ranking
import numpy as np
from itertools import product
import copy
import math

def ranks_above(ranking,c):
    """
    Returns the number of positions above candidate ``c`` in ``ranking``, taking into account ties in Ranking objects.
    """
    if isinstance(ranking, tuple):

        return ranking.index(c)

    if isinstance(ranking, Ranking):

        ranking.normalize_ranks()

        if any([ranking.rmap[d] == ranking.rmap[c] for d in ranking.cands if c!=d]):
            return 2 * (ranking.rmap[c] - 1) + 1
        else:
            return 2 * (ranking.rmap[c] - 1)
    
def ranks_below(ranking,c):
    """
    Returns the number of positions below candidate ``c`` in ``ranking``, taking into account ties in Ranking objects.
    """
    if isinstance(ranking, tuple):
            
        return len(ranking) - ranking.index(c) - 1
    
    if isinstance(ranking, Ranking):

        ranking.normalize_ranks()

        if any([ranking.rmap[d] == ranking.rmap[c] for d in ranking.cands if c!=d]):
            return 2 * (len(ranking.cands) - ranking.rmap[c]) + 1
        else:
            return 2 * (len(ranking.cands) - ranking.rmap[c])

def n_rank_lift(ranking, c, n):
    """
    Return a ranking in which ``c`` is moved up n positions in ``ranking``.
    """
    if isinstance(ranking, tuple):
        assert c not in ranking[:n], f"there are not enough ranks above {c} to lift {c} {n} ranks"
        _new_ranking = copy.deepcopy(ranking)
        c_idx = _new_ranking.index(c)
        new_ranking = _new_ranking[:c_idx-n] + (_new_ranking[c_idx],) + _new_ranking[c_idx-n:c_idx] + _new_ranking[c_idx+1:]
    
    if isinstance(ranking, Ranking):

        ranking.normalize_ranks()

        assert ranks_above(ranking,c) >= n, f"there are not enough ranks above {c} to lift {c} {n} ranks"

        new_ranking_dict = dict()

        if any([ranking.rmap[d] == ranking.rmap[c] for d in ranking.cands if c!=d]): # if c is tied with another candidate

            if n%2 == 0: 
                for d in ranking.cands:
                    if d == c:
                        new_ranking_dict[d] = ranking.rmap[d] - (n/2)
                    else:
                        new_ranking_dict[d] = ranking.rmap[d]

            else:
                for d in ranking.cands:
                    if d == c:
                        new_ranking_dict[d] = ranking.rmap[d] - (math.floor(n/2) + math.ceil(n/2))/2
                    else:
                        new_ranking_dict[d] = ranking.rmap[d]
                
            new_ranking = Ranking(new_ranking_dict)

        else:
            if n%2 == 0:
                for d in ranking.cands:
                    if d == c:
                        new_ranking_dict[d] = ranking.rmap[d] - ((n//2) + (n//2 + 1))/2
                    else:
                        new_ranking_dict[d] = ranking.rmap[d]
            else:
                for d in ranking.cands:
                    if d == c:
                        new_ranking_dict[d] = ranking.rmap[d] - math.ceil(n/2)
                    else:
                        new_ranking_dict[d] = ranking.rmap[d]

            new_ranking = Ranking(new_ranking_dict)

        new_ranking.normalize_ranks()

    return new_ranking

def n_rank_drop(ranking, c, n):
    """
    Return a ranking in which ``c`` is moved down n positions in ``ranking``.
    """
    if isinstance(ranking, tuple):
        # assert that there are n ranks below c
        assert len(ranking) - ranking.index(c) - 1 >= n, f"there are not enough ranks below {c} to drop {c} {n} ranks"
        _new_ranking = copy.deepcopy(ranking)
        c_idx = _new_ranking.index(c)
        new_ranking = _new_ranking[:c_idx] + _new_ranking[c_idx+1:c_idx+n+1] + (_new_ranking[c_idx],) + _new_ranking[c_idx+n+1:]
    
    if isinstance(ranking, Ranking):

        ranking.normalize_ranks()

        assert ranks_below(ranking,c) >= n, f"there are not enough ranks below {c} to drop {c} {n} ranks"

        new_ranking_dict = dict()

        if any([ranking.rmap[d] == ranking.rmap[c] for d in ranking.cands if c!=d]): # if c is tied with another candidate

            if n%2 == 0: 
                for d in ranking.cands:
                    if d == c:
                        new_ranking_dict[d] = ranking.rmap[d] + (n/2)
                    else:
                        new_ranking_dict[d] = ranking.rmap[d]

            else:
                for d in ranking.cands:
                    if d == c:
                        new_ranking_dict[d] = ranking.rmap[d] + (math.floor(n/2) + math.ceil(n/2))/2
                    else:
                        new_ranking_dict[d] = ranking.rmap[d]
                
            new_ranking = Ranking(new_ranking_dict)

        else:
            if n%2 == 0:
                for d in ranking.cands:
                    if d == c:
                        new_ranking_dict[d] = ranking.rmap[d] + ((n//2) + (n//2 + 1))/2
                    else:
                        new_ranking_dict[d] = ranking.rmap[d]
            else:
                for d in ranking.cands:
                    if d == c:
                        new_ranking_dict[d] = ranking.rmap[d] + math.ceil(n/2)
                    else:
                        new_ranking_dict[d] = ranking.rmap[d]

            new_ranking = Ranking(new_ranking_dict)

        new_ranking.normalize_ranks()
    
    return new_ranking

def has_monotonicity_violation(profile, vm, verbose = False, violation_type = "Lift", check_probabilities = False, one_rank_monotonicity = False): 
    """
    If violation_type = "Lift", returns True if there is some winning candidate A and some voter v such that lifting A up some number of positions in v's ranking causes A to lose.

    If violation_type = "Drop", returns True if there is some losing candidate A and some voter v such that dropping A down some number of positions in v's ranking causes A to win.

    If checking_probabilities = True, returns True if there is some candidate whose probability of winning decreases after a lifting or increases after a dropping.

    If one_rank_monotonicity = True, then the function will check lifts/drops of one rank only.
    
    Args:
        profile: a Profile object.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 
        violation_type: default is "Lift"

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    .. note::
        If a voting method violates monotonicity, then it violates one-rank monotonicity, so setting one_rank_monotonicity = True is sufficient for testing whether a method violates monotonicity (though not for testing the frequency of monotonicity violations).

    """
    
    _rankings, _rcounts = profile.rankings_counts

    if isinstance(profile, Profile):
        rankings = [tuple(r) for r in list(_rankings)]

    if isinstance(profile, ProfileWithTies):
        rankings = _rankings

    rcounts = list(_rcounts)
    old_rankings = copy.deepcopy(rankings)

    ws = vm(profile)

    if violation_type == "Lift":
        for w in ws: 
            for r_idx, r in enumerate(rankings):

                if rcounts[r_idx] == 0:
                    continue

                if isinstance(r, Ranking): # Make sure all candidates are ranked in r
                    r = Ranking({a: r.rmap[a] if a in r.cands else max(r.ranks)+1 for a in profile.candidates})

                if r[0] != w:
                    old_ranking = copy.deepcopy(r)

                    if one_rank_monotonicity:
                        ranks_above_w = 1
                    else:
                        ranks_above_w = ranks_above(r, w)

                    for n in range(1, ranks_above_w+1):

                        new_ranking = n_rank_lift(r, w, n)
                        new_rankings = old_rankings + [new_ranking]
                        new_rcounts  = copy.deepcopy(rcounts + [1])
                        new_rcounts[r_idx] -= 1

                        if isinstance(profile, Profile):
                            new_prof = Profile(new_rankings, new_rcounts)
                        if isinstance(profile, ProfileWithTies):
                            new_prof = ProfileWithTies(new_rankings, new_rcounts)
                            if profile.using_extended_strict_preference:
                                new_prof.use_extended_strict_preference()

                        new_ws = vm(new_prof)
                        
                        if w not in new_ws: 
                            if verbose: 
                                if n==1:
                                    print(f"Monotonicity violation for {vm.name} by lifting {w} one rank:")
                                else:
                                    print(f"Monotonicity violation for {vm.name} by lifting {w} by {n} ranks:")
                                profile.display()
                                print(profile.description())
                                profile.display_margin_graph()
                                print(f"{vm.name} winners: ", ws)
                                print("Original ranking: ", old_ranking)
                                print(f"New ranking: {new_ranking}")
                                new_prof.display()
                                print(new_prof.description())
                                new_prof.display_margin_graph()
                                print(f"{vm.name} winners in updated profile:", new_ws)
                            return True
                        
                        if w in new_ws and check_probabilities == True and len(new_ws) > len(ws):
                            if verbose: 
                                if n==1:
                                    print(f"Probabilistic monotonicity violation for {vm.name} by lifting {w} one rank:")
                                else:
                                    print(f"Probabilistic monotonicity violation for {vm.name} by lifting {w} by {n} ranks:")
                                profile.display()
                                print(profile.description())
                                profile.display_margin_graph()
                                print(f"{vm.name} winners: ", ws)
                                print("Original ranking: ", old_ranking)
                                print(f"New ranking: {new_ranking}")
                                new_prof.display()
                                print(new_prof.description())
                                new_prof.display_margin_graph()
                                print(f"{vm.name} winners in updated profile:", new_ws)
                            return True
                    
    elif violation_type == "Drop":
        for l in profile.candidates:
            if l not in ws:
                for r_idx, r in enumerate(rankings): 

                    if rcounts[r_idx] == 0:
                        continue

                    if isinstance(r, Ranking): # Make sure all candidates are ranked in r
                        r = Ranking({a: r.rmap[a] if a in r.cands else max(r.ranks)+1 for a in profile.candidates})

                    if r[-1] != l:
                        old_ranking = copy.deepcopy(r)

                        if one_rank_monotonicity:
                            ranks_below_l = 1
                        else:
                            ranks_below_l = ranks_below(r, l)

                        for n in range(1, ranks_below_l+1):
                            
                            new_ranking = n_rank_drop(r, l, n)
                            new_rankings = old_rankings + [new_ranking]
                            new_rcounts  = copy.deepcopy(rcounts + [1])
                            new_rcounts[r_idx] -= 1

                            if isinstance(profile, Profile):
                                new_prof = Profile(new_rankings, new_rcounts)
                            if isinstance(profile, ProfileWithTies):
                                new_prof = ProfileWithTies(new_rankings, new_rcounts)
                                if profile.using_extended_strict_preference:
                                    new_prof.use_extended_strict_preference()

                            new_ws = vm(new_prof)

                            if l in new_ws: 
                                if verbose: 
                                    if n==1:
                                        print(f"Monotonicity violation for {vm.name} by dropping {l} one rank:")
                                    else:
                                        print(f"Monotonicity violation for {vm.name} by dropping {l} by {n} ranks:")
                                    profile.display()
                                    print(profile.description())
                                    profile.display_margin_graph()
                                    print(f"{vm.name} winners: ", ws)
                                    print("Original ranking: ", old_ranking)
                                    print(f"New ranking: {new_ranking}")
                                    new_prof.display()
                                    print(new_prof.description())
                                    new_prof.display_margin_graph()
                                    print(f"{vm.name} winners in updated profile: ", new_ws)
                                return True
                            
            if check_probabilities and l in ws:
                for r_idx, r in enumerate(rankings): 

                    if rcounts[r_idx] == 0:
                        continue

                    if isinstance(r, Ranking): # Make sure all candidates are ranked in r
                        r = Ranking({a: r.rmap[a] if a in r.cands else max(r.ranks)+1 for a in profile.candidates})

                    if r[-1] != l:
                        old_ranking = copy.deepcopy(r)

                        if one_rank_monotonicity:
                            ranks_below_l = 1
                        else:
                            ranks_below_l = ranks_below(r, l)

                        new_ranking = one_rank_drop(r, l)
                        new_rankings = old_rankings + [new_ranking]
                        new_rcounts  = copy.deepcopy(rcounts + [1])
                        new_rcounts[r_idx] -= 1

                        if isinstance(profile, Profile):
                            new_prof = Profile(new_rankings, new_rcounts)
                        if isinstance(profile, ProfileWithTies):
                            new_prof = ProfileWithTies(new_rankings, new_rcounts)
                            if profile.using_extended_strict_preference:
                                new_prof.use_extended_strict_preference()

                        new_ws = vm(new_prof)
                        if l in new_ws and len(new_ws) < len(ws): 
                            if verbose: 
                                if n==1:
                                    print(f"Probabilistic monotonicity violation for {vm.name} by dropping {l} one rank:")
                                else:
                                    print(f"Probabilistic monotonicity violation for {vm.name} by dropping {l} by {n} ranks:")
                                profile.display()
                                print(profile.description())
                                profile.display_margin_graph()
                                print(f"{vm.name} winners: ", ws)
                                print("Original ranking: ", old_ranking)
                                print(f"New ranking: {new_ranking}")
                                new_prof.display()
                                print(new_prof.description())
                                new_prof.display_margin_graph()
                                print(f"{vm.name} winners in updated profile: ", new_ws)
                            return True

    return False

def find_all_monotonicity_violations(profile, vm, verbose = False, violation_type = "Lift", check_probabilities = False, one_rank_monotonicity = False):
    """
    If violation_type = "Lift", returns all tuples (candidate, ranking, "Lift", n) such that the candidate wins in the original profile but loses after lifting the candidate up n positions in the ranking.

    If violation_type = "Drop", returns all tuples (candidate, ranking, "Drop", n) such that the candidate loses in the original profile but wins after dropping the candidate down n positions in the ranking.

    If checking_probabilities = True, returns all tuples (candidate, ranking, violation_type, n) such that the candidate's probability of winning decreases after a lifting or increases after a dropping.

    If one_rank_monotonicity = True, then the function will check lifts/drops of one rank only.

    Args:
        profile: a Profile object.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 
        violation_type: default is "Lift"

    Returns:
        A list of tuples (candidate, ranking, violation_type, positions lifted/dropped) witnessing violations of monotonicity.

    .. note::
        If a voting method violates monotonicity, then it violates one-rank monotonicity, so setting one_rank_monotonicity = True is sufficient for testing whether a method violates monotonicity (though not for testing the frequency of monotonicity violations).
    """

    _rankings, _rcounts = profile.rankings_counts

    if isinstance(profile, Profile):
        rankings = [tuple(r) for r in list(_rankings)]

    if isinstance(profile, ProfileWithTies):
        rankings = _rankings

    rcounts = list(_rcounts)
    old_rankings = copy.deepcopy(rankings)

    ws = vm(profile)
    witnesses = list()

    if violation_type == "Lift":
        for w in ws: 
            for r_idx, r in enumerate(rankings): 

                if rcounts[r_idx] == 0:
                    continue

                if isinstance(r, Ranking): # Make sure all candidates are ranked in r
                    r = Ranking({a: r.rmap[a] if a in r.cands else max(r.ranks)+1 for a in profile.candidates})
                    
                if r[0] != w:
                    old_ranking = copy.deepcopy(r)
                    
                    if one_rank_monotonicity:
                        ranks_above_w = 1
                    else:
                        ranks_above_w = ranks_above(r, w)

                    for n in range(1, ranks_above_w+1):
                        new_ranking = n_rank_lift(r, w, n)
                        new_rankings = old_rankings + [new_ranking]
                        new_rcounts  = copy.deepcopy(rcounts + [1])
                        new_rcounts[r_idx] -= 1

                        if isinstance(profile, Profile):
                            new_prof = Profile(new_rankings, new_rcounts)
                        if isinstance(profile, ProfileWithTies):
                            new_prof = ProfileWithTies(new_rankings, new_rcounts)
                            if profile.using_extended_strict_preference:
                                new_prof.use_extended_strict_preference()

                        new_ws = vm(new_prof)

                        if w not in new_ws: 
                            witnesses.append((w, old_ranking, "Lift", n))
                            if verbose: 
                                if n==1:
                                    print(f"Monotonicity violation for {vm.name} by lifting {w} one rank:")
                                else:
                                    print(f"Monotonicity violation for {vm.name} by lifting {w} by {n} ranks:")
                                profile.display()
                                print(profile.description())
                                profile.display_margin_graph()
                                print(f"{vm.name} winners: ", ws)
                                print("Original ranking ", old_ranking)
                                print(f"New ranking: {new_ranking}")
                                new_prof.display()
                                print(new_prof.description())
                                new_prof.display_margin_graph()
                                print(f"{vm.name} winners in updated profile: ", new_ws)
                                print("")

                        if w in new_ws and check_probabilities == True and len(new_ws) > len(ws):
                            witnesses.append((w, old_ranking, "Lift", n))
                            if verbose: 
                                if n==1:
                                    print(f"Probabilistic monotonicity violation for {vm.name} by lifting {w} one rank:")
                                else:
                                    print(f"Probabilistic monotonicity violation for {vm.name} by lifting {w} by {n} ranks:")
                                profile.display()
                                print(profile.description())
                                profile.display_margin_graph()
                                print(f"{vm.name} winners: ", ws)
                                print("Original ranking: ", old_ranking)
                                print(f"New ranking: {new_ranking}")
                                new_prof.display()
                                print(new_prof.description())
                                new_prof.display_margin_graph()
                                print(f"{vm.name} winners in updated profile:", new_ws)
                                print("")

    elif violation_type == "Drop":
        for l in profile.candidates:
            if l not in ws:
                for r_idx, r in enumerate(rankings): 

                    if rcounts[r_idx] == 0:
                        continue

                    if isinstance(r, Ranking): # Make sure all candidates are ranked in r
                        r = Ranking({a: r.rmap[a] if a in r.cands else max(r.ranks)+1 for a in profile.candidates})

                    if r[-1] != l:
                        old_ranking = copy.deepcopy(r)
                        
                        if one_rank_monotonicity:
                            ranks_below_l = 1
                        else:
                            ranks_below_l = ranks_below(r, l)

                        for n in range(1, ranks_below_l+1):
                            new_ranking = n_rank_drop(r, l, n)
                            new_rankings = old_rankings + [new_ranking]
                            new_rcounts  = copy.deepcopy(rcounts + [1])
                            new_rcounts[r_idx] -= 1

                            if isinstance(profile, Profile):
                                new_prof = Profile(new_rankings, new_rcounts)
                            if isinstance(profile, ProfileWithTies):
                                new_prof = ProfileWithTies(new_rankings, new_rcounts)
                                if profile.using_extended_strict_preference:
                                    new_prof.use_extended_strict_preference()

                            new_ws = vm(new_prof)

                            if l in new_ws: 
                                witnesses.append((l, old_ranking, "Drop", n))
                                if verbose: 
                                    if n==1:
                                        print(f"Monotonicity violation for {vm.name} by dropping {l} one rank:")
                                    else:
                                        print(f"Monotonicity violation for {vm.name} by dropping {l} by {n} ranks:")
                                    profile.display()
                                    print(profile.description())
                                    profile.display_margin_graph()
                                    print(f"{vm.name} winners: ", ws)
                                    print("Original ranking: ", old_ranking)
                                    print(f"New ranking: {new_ranking}")
                                    new_prof.display()
                                    print(new_prof.description())
                                    new_prof.display_margin_graph()
                                    print(f"{vm.name} winners in updated profile: ", new_ws)
                                    print("")

            if check_probabilities and l in ws:
                for r_idx, r in enumerate(rankings): 

                    if rcounts[r_idx] == 0:
                        continue

                    if isinstance(r, Ranking): # Make sure all candidates are ranked in r
                        r = Ranking({a: r.rmap[a] if a in r.cands else max(r.ranks)+1 for a in profile.candidates})

                    if r[-1] != l:
                        old_ranking = copy.deepcopy(r)
                        
                        if one_rank_monotonicity:
                            ranks_below_l = 1
                        else:
                            ranks_below_l = ranks_below(r, l)

                        for n in range(1, ranks_below_l+1):
                            new_ranking = n_rank_drop(r, l, n)
                            new_rankings = old_rankings + [new_ranking]
                            new_rcounts  = copy.deepcopy(rcounts + [1])
                            new_rcounts[r_idx] -= 1

                            if isinstance(profile, Profile):
                                new_prof = Profile(new_rankings, new_rcounts)
                            if isinstance(profile, ProfileWithTies):
                                new_prof = ProfileWithTies(new_rankings, new_rcounts)
                                if profile.using_extended_strict_preference:
                                    new_prof.use_extended_strict_preference()

                            new_ws = vm(new_prof)

                            if l in new_ws and len(new_ws) < len(ws): 
                                witnesses.append((l, old_ranking, "Drop", n))
                                if verbose: 
                                    if n==1:
                                        print(f"Probabilistic monotonicity violation for {vm.name} by dropping {l} one rank:")
                                    else:
                                        print(f"Probabilistic monotonicity violation for {vm.name} by dropping {l} by {n} ranks:")
                                    profile.display()
                                    print(profile.description())
                                    profile.display_margin_graph()
                                    print(f"{vm.name} winners: ", ws)
                                    print("Original ranking: ", old_ranking)
                                    print(f"New ranking: {new_ranking}")
                                    new_prof.display()
                                    print(new_prof.description())
                                    new_prof.display_margin_graph()
                                    print(f"{vm.name} winners in updated profile: ", new_ws)
                                    print("")

    return witnesses
                    
monotonicity = Axiom(
    "Monotonicity",
    has_violation = has_monotonicity_violation,
    find_all_violations = find_all_monotonicity_violations,
)

def lift_to_first(ranking, c):
    """
    Return a ranking in which ``c`` is moved to first position in ``ranking``.
    """

    if isinstance(ranking, tuple):
        assert c != ranking[0], "can't lift a candidate already in first place"
        new_ranking = copy.deepcopy(ranking)
        c_idx = new_ranking.index(c)
        new_ranking = (c,) + new_ranking[:c_idx] + new_ranking[c_idx+1:]

    if isinstance(ranking, Ranking):
        assert not ranking.first() == [c], "can't lift a candidate already uniquely in first place"
        new_ranking = Ranking({a: ranking.rmap[a] if a !=c else min(ranking.ranks) - 1 for a in ranking.cands})
        new_ranking.normalize_ranks()

    return new_ranking

def drop_to_last(ranking, c):
    """
    Return a ranking in which ``c`` is moved to last position in ``ranking``.
    """

    if isinstance(ranking, tuple):
        assert c != ranking[-1], "can't drop a candidate already in last place"
        new_ranking = copy.deepcopy(ranking)
        c_idx = new_ranking.index(c)
        new_ranking = new_ranking[:c_idx] + new_ranking[c_idx+1:] + (c,)

    if isinstance(ranking, Ranking):
        assert not ranking.last() == [c], "can't drop a candidate already uniquely in last place"
        new_ranking = Ranking({a: ranking.rmap[a] if a !=c else max(ranking.ranks) + 1 for a in ranking.cands})
        new_ranking.normalize_ranks()

    return new_ranking

def has_weak_positive_responsiveness_violation(profile, vm, verbose = False, violation_type="Lift"): 
    """
    If violation_type = "Lift", returns True if there is some winning candidate A and some voter v who ranks A last such that v moving A into first place does not make A the unique winner

    If violation_type = "Drop", returns True if there is some candidate A who is either a loser or a non-unique winner and some voter v who ranks A first such that v moving A into last place does not make A a loser.
    
    Args:
        profile: a Profile object.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 
        violation_type: default is "Lift"

    Returns: 
        Result of the test (bool): Returns True if there is a violation and False otherwise. 

    """
    
    _rankings, _rcounts = profile.rankings_counts

    if isinstance(profile, Profile):
        rankings = [tuple(r) for r in list(_rankings)]
    
    if isinstance(profile, ProfileWithTies):
        rankings = _rankings

    rcounts = list(_rcounts)
    old_rankings = copy.deepcopy(rankings)

    ws = vm(profile)

    if violation_type == "Lift":
        for w in ws: 
            for r_idx, r in enumerate(rankings): 

                if rcounts[r_idx] == 0:
                    continue

                if isinstance(r, Ranking): # Make sure all candidates are ranked in r
                    r = Ranking({a: r.rmap[a] if a in r.cands else max(r.ranks)+1 for a in profile.candidates})

                if r[-1] == w and not r[0] == w:
                    old_ranking = copy.deepcopy(r)
                    new_ranking = lift_to_first(r, w)
                    new_rankings = old_rankings + [new_ranking]
                    new_rcounts  = copy.deepcopy(rcounts + [1])
                    new_rcounts[r_idx] -= 1

                    if isinstance(profile, Profile):
                        new_prof = Profile(new_rankings, new_rcounts)
                    
                    if isinstance(profile, ProfileWithTies):
                        new_prof = ProfileWithTies(new_rankings, new_rcounts)
                        if profile.using_extended_strict_preference:
                            new_prof.use_extended_strict_preference()

                    new_ws = vm(new_prof)

                    if len(new_ws) > 1 or (len(new_ws) == 1 and new_ws[0] != w):
                        if verbose: 
                            print(f"Weak positive responsiveness violation for {vm.name} by lifting {w}:")
                            profile.display()
                            print(profile.description())
                            profile.display_margin_graph()
                            print(f"{vm.name} winners: ", ws)
                            print("Original ranking: ", old_ranking)
                            print(f"New ranking: {new_ranking}")
                            new_prof.display()
                            print(new_prof.description())
                            new_prof.display_margin_graph()
                            print(f"{vm.name} winners in updated profile:", new_ws)
                        return True
                    
    elif violation_type == "Drop":
        for l in profile.candidates:
            if l not in ws or (l in ws and len(ws) > 1):
                for r_idx, r in enumerate(rankings): 

                    if rcounts[r_idx] == 0:
                        continue

                    if isinstance(r, Ranking): # Make sure all candidates are ranked in r
                        r = Ranking({a: r.rmap[a] if a in r.cands else max(r.ranks)+1 for a in profile.candidates})

                    if r[0] == l and not r[-1] == l:
                        old_ranking = copy.deepcopy(r)
                        new_ranking = drop_to_last(r, l)
                        new_rankings = old_rankings + [new_ranking]
                        new_rcounts  = copy.deepcopy(rcounts + [1])
                        new_rcounts[r_idx] -= 1

                        if isinstance(profile, Profile):
                            new_prof = Profile(new_rankings, new_rcounts)
                        if isinstance(profile, ProfileWithTies):
                            new_prof = ProfileWithTies(new_rankings, new_rcounts)
                            if profile.using_extended_strict_preference:
                                new_prof.use_extended_strict_preference()

                        new_ws = vm(new_prof)

                        if l in new_ws: 
                            if verbose: 
                                print(f"Weak positive responsiveness violation for {vm.name} by dropping {l}:")
                                profile.display()
                                print(profile.description())
                                profile.display_margin_graph()
                                print(f"{vm.name} winners: ", ws)
                                print("Original ranking: ", old_ranking)
                                print(f"New ranking: {new_ranking}")
                                new_prof.display()
                                print(new_prof.description())
                                new_prof.display_margin_graph()
                                print(f"{vm.name} winners in updated profile: ", new_ws)
                            return True

    return False

def find_all_weak_positive_responsiveness_violations(profile, vm, verbose = False, violation_type="Lift"):
    """
    If violation_type = "Lift", returns all pairs (candidate, ranking) such that the candidate is a unique winner in the original profile but is not a unique winner after the voter moves the candidate from last to first place in the ranking.

    If violation_type = "Drop", returns all pairs (candidate, ranking) such that the candidate is either a loser or a non-unique winner in the original profile but is a  winner after the voter moves the candidate from first to last place in the ranking.

    Args:
        profile: a Profile object.
        vm (VotingMethod): A voting method to test.
        verbose (bool, default=False): If a violation is found, display the violation. 
        violation_type: default is "Lift"

    Returns:
        A list of pairs (candidate, ranking) witnessing violations of weak positive responsiveness.

    """

    _rankings, _rcounts = profile.rankings_counts

    if isinstance(profile, Profile):
        rankings = [tuple(r) for r in list(_rankings)]

    if isinstance(profile, ProfileWithTies):
        rankings = _rankings

    rcounts = list(_rcounts)
    old_rankings = copy.deepcopy(rankings)

    ws = vm(profile)
    witnesses = list()

    if violation_type == "Lift":
        for w in ws: 
            for r_idx, r in enumerate(rankings): 

                if rcounts[r_idx] == 0:
                    continue

                if isinstance(r, Ranking): # Make sure all candidates are ranked in r
                    r = Ranking({a: r.rmap[a] if a in r.cands else max(r.ranks)+1 for a in profile.candidates})

                if r[-1] == w and not r[0] == w:
                    old_ranking = copy.deepcopy(r)
                    new_ranking = lift_to_first(r, w)
                    new_rankings = old_rankings + [new_ranking]
                    new_rcounts  = copy.deepcopy(rcounts + [1])
                    new_rcounts[r_idx] -= 1

                    if isinstance(profile, Profile):
                        new_prof = Profile(new_rankings, new_rcounts)
                    if isinstance(profile, ProfileWithTies):
                        new_prof = ProfileWithTies(new_rankings, new_rcounts)
                        if profile.using_extended_strict_preference:
                            new_prof.use_extended_strict_preference()

                    new_ws = vm(new_prof)

                    if len(new_ws) > 1 or (len(new_ws) == 1 and new_ws[0] != w):
                        witnesses.append((w, old_ranking, "Lift"))
                        if verbose: 
                            print(f"Weak positive responsiveness violation for {vm.name} by lifting {w}:")
                            profile.display()
                            print(profile.description())
                            profile.display_margin_graph()
                            print(f"{vm.name} winners: ", ws)
                            print("Original ranking ", old_ranking)
                            print(f"New ranking: {new_ranking}")
                            new_prof.display()
                            print(new_prof.description())
                            new_prof.display_margin_graph()
                            print(f"{vm.name} winners in updated profile: ", new_ws)

    elif violation_type == "Drop":
        for l in profile.candidates:
            if l not in ws or (l in ws and len(ws) > 1):
                for r_idx, r in enumerate(rankings): 

                    if rcounts[r_idx] == 0:
                        continue

                    if isinstance(r, Ranking): # Make sure all candidates are ranked in r
                        r = Ranking({a: r.rmap[a] if a in r.cands else max(r.ranks)+1 for a in profile.candidates})

                    if r[0] == l and not r[-1] == l:
                        old_ranking = copy.deepcopy(r)
                        new_ranking = drop_to_last(r, l)
                        new_rankings = old_rankings + [new_ranking]
                        new_rcounts  = copy.deepcopy(rcounts + [1])
                        new_rcounts[r_idx] -= 1

                        if isinstance(profile, Profile):
                            new_prof = Profile(new_rankings, new_rcounts)

                        if isinstance(profile, ProfileWithTies):
                            new_prof = ProfileWithTies(new_rankings, new_rcounts)
                            if profile.using_extended_strict_preference:
                                new_prof.use_extended_strict_preference()

                        new_ws = vm(new_prof)

                        if l in new_ws: 
                            witnesses.append((l, old_ranking, "Drop"))
                            if verbose: 
                                print(f"Weak positive responsiveness violation for {vm.name} by dropping {l}:")
                                profile.display()
                                print(profile.description())
                                profile.display_margin_graph()
                                print(f"{vm.name} winners: ", ws)
                                print("Original ranking: ", old_ranking)
                                print(f"New ranking: {new_ranking}")
                                new_prof.display()
                                print(new_prof.description())
                                new_prof.display_margin_graph()
                                print(f"{vm.name} winners in updated profile: ", new_ws)
    
    return witnesses

weak_positive_responsiveness = Axiom(
    "Weak Positive Responsiveness",
    has_violation = has_weak_positive_responsiveness_violation,
    find_all_violations = find_all_weak_positive_responsiveness_violations,
)
                   
monotonicity_axioms = [
    monotonicity,
    weak_positive_responsiveness
]


"""
    File: profiles_with_ties.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: January 5, 2022

    A class that represents profiles of (truncated) strict weak orders.
"""

from math import ceil
import copy
import numpy as np
from tabulate import tabulate
from pref_voting.profiles import Profile
from pref_voting.rankings import Ranking
from pref_voting.scoring_methods import symmetric_borda_scores
from pref_voting.weighted_majority_graphs import (
    MajorityGraph,
    MarginGraph,
    SupportGraph,
)
import os
import pandas as pd

def _num_rank_profile_with_ties(rankings, rcounts, cand, level):
    """
    Counts the number of voters that rank candidate `cand` at rank `level` (1-based)
    in a ProfileWithTies object.

    Args:
        rankings: list of Ranking objects
        rcounts: list of counts for each ranking
        cand: candidate
        level: rank to check (1-based)

    Returns:
        Total number of voters ranking candidate `cand` at rank `level`
    """
    total = 0
    for ranking, count in zip(rankings, rcounts):
        if cand in ranking.rmap and ranking.rmap[cand] == level - 1:  # Convert 1-based rank to 0-based
            total += count
    return total

def same_ranking_extended_strict_pref(ranking1, ranking2, candidates): 
    # check if ranking1 and ranking2 have the same ranking of candidates
    for c1 in candidates:
        for c2 in candidates:
            if (not ranking1.extended_strict_pref(c1, c2) and ranking2.extended_strict_pref(c1, c2)) or (not ranking2.extended_strict_pref(c1, c2) and ranking1.extended_strict_pref(c1, c2)):
                return False
    return True

class ProfileWithTies(object):
    """An anonymous profile of (truncated) strict weak orders of :math:`n` candidates. 

    :param rankings: List of rankings in the profile, where a ranking is either a :class:`Ranking` object or a dictionary.
    :type rankings: list[dict[int or str: int]] or list[Ranking]
    :param rcounts: List of the number of voters associated with each ranking.  Should be the same length as rankings.  If not provided, it is assumed that 1 voters submitted each element of ``rankings``.
    :type rcounts: list[int], optional
    :param candidates: List of candidates in the profile.  If not provided, this is the list that is ranked by at least on voter.
    :type candidates: list[int] or list[str], optional
    :param cmap: Dictionary mapping candidates (integers) to candidate names (strings).  If not provided, each candidate name is mapped to itself.
    :type cmap: dict[int: str], optional

    :Example:

    The following code creates a profile in which
    2 voters submitted the ranking 0 ranked first, 1 ranked second, and 2 ranked third; 3 voters submitted the ranking 1 and 2 are tied for first place and 0 is ranked second; and 1 voter submitted the ranking in which 2 is ranked first and 0 is ranked second:

    .. code-block:: python

            prof = ProfileWithTies([{0: 1, 1: 2, 2: 3}, {1:1, 2:1, 0:2}, {2:1, 0:2}], [2, 3, 1])
    """

    def __init__(self, rankings, rcounts=None, candidates=None, cmap=None):
        """constructor method"""

        assert rcounts is None or len(rankings) == len(
            rcounts
        ), "The number of rankings much be the same as the number of rcounts"
        

        get_cands = lambda r: list(r.keys()) if type(r) == dict else r.cands
        self.candidates = (
            sorted(candidates)
            if candidates is not None
            else sorted(list(set([c for r in rankings for c in get_cands(r)])))
        )
        """The candidates in the profile. """

        self.num_cands = len(self.candidates)
        """The number of candidates in the profile."""

        self.cmap = cmap if cmap is not None else {c: str(c) for c in self.candidates}
        """The candidate map is a dictionary associating a candidate with the name used when displaying a candidate."""

        self._rankings = [
            Ranking(r, cmap=self.cmap)
            if type(r) == dict
            else Ranking(r.rmap, cmap=self.cmap)
            for r in rankings
        ]
        """The list of rankings in the Profile (each ranking is a :class:`Ranking` object). 
        """

        self.ranks = list(range(1, self.num_cands + 1))
        """The ranks that are possible in the profile. """

        self.cindices = list(range(self.num_cands))
        self._cand_to_cindex = {c: i for i, c in enumerate(self.candidates)}
        self.cand_to_cindex = lambda c: self._cand_to_cindex[c]
        self._cindex_to_cand = {i: c for i, c in enumerate(self.candidates)}
        self.cindex_to_cand = lambda i: self._cindex_to_cand[i]
        """Maps candidates to their index in the list of candidates and vice versa. """
    
        self.rcounts = [1] * len(rankings) if rcounts is None else list(rcounts)

        self.num_voters = np.sum(self.rcounts)
        """The number of voters in the profile. """

        self.using_extended_strict_preference = False
        """A flag indicating whether the profile is using extended strict preferences when calculating supports, margins, etc."""
        
        # memoize the supports
        self._supports = {
            c1: {
                c2: sum(
                    n
                    for r, n in zip(self._rankings, self.rcounts)
                    if r.strict_pref(c1, c2)
                )
                for c2 in self.candidates
            }
            for c1 in self.candidates
        }

    def use_extended_strict_preference(self):
        """
        Redefine the supports so that *extended strict preferences* are used. Using extended strict preference may change the margins between candidates.
        """

        self.using_extended_strict_preference = True
        self._supports = {
            c1: {
                c2: sum(
                    n
                    for r, n in zip(self._rankings, self.rcounts)
                    if r.extended_strict_pref(c1, c2)
                )
                for c2 in self.candidates
            }
            for c1 in self.candidates
        }

    def use_strict_preference(self):
        """
        Redefine the supports so that strict preferences are used. Using strict preference may change the margins between candidates.
        """

        self.using_extended_strict_preference = False
        self._supports = {
            c1: {
                c2: sum(
                    n
                    for r, n in zip(self._rankings, self.rcounts)
                    if r.strict_pref(c1, c2)
                )
                for c2 in self.candidates
            }
            for c1 in self.candidates
        }
    @property 
    def rankings(self): 
        """
        Return a list of all individual rankings in the profile. 
        """
        
        return [r for ridx,r in enumerate(self._rankings) 
                for _ in range(self.rcounts[ridx])]

    @property 
    def rankings_as_indifference_list(self): 
        """
        Return a list of all individual rankings as indifference lists in the profile. 
        """
        
        return [r.to_indiff_list() for ridx,r in enumerate(self._rankings) 
                for _ in range(self.rcounts[ridx])]

    @property 
    def ranking_types(self): 
        """
        Return a list of the types of rankings in the profile. 
        """
        
        unique_rankings = []
        for r in self._rankings: 
            if r not in unique_rankings: 
                unique_rankings.append(r)
        return unique_rankings
    
    @property
    def rankings_counts(self):
        """
        Returns the rankings and the counts of each ranking.
        """

        return self._rankings, self.rcounts

    @property
    def rankings_as_dicts_counts(self):
        """
        Returns the rankings represented as dictionaries and the counts of each ranking.
        """

        return [r.rmap for r in self._rankings], self.rcounts

    def support(self, c1, c2):
        """
        Returns the support of candidate ``c1`` over candidate ``c2``, where the support is the number of voters that rank ``c1`` strictly above ``c2``.
        """

        return self._supports[c1][c2]

    def margin(self, c1, c2):
        """
        Returns the margin of candidate ``c1`` over candidate ``c2``, where the margin is the number of voters that rank ``c1`` strictly above ``c2`` minus the number of voters that rank ``c2`` strictly above ``c1``.
        """

        return self._supports[c1][c2] - self._supports[c2][c1]

    @property
    def margin_matrix(self):
        """Returns the margin matrix of the profile, where the entry at row ``i`` and column ``j`` is the margin of candidate ``i`` over candidate ``j``."""

        return np.array(
            [[self.margin(self.cindex_to_cand(c1_idx), self.cindex_to_cand(c2_idx)) for c2_idx in self.cindices] for c1_idx in self.cindices]
        )
    
    def is_tied(self, c1, c2): 
        """Returns True if ``c1`` and ``c2`` are tied (i.e., the margin of ``c1`` over ``c2`` is 0)."""

        return self.margin(c1, c2) == 0

    def dominators(self, cand, curr_cands=None):
        """
        Returns the list of candidates that are majority preferred to ``cand`` in the profile restricted to the candidates in ``curr_cands``.
        """
        candidates = self.candidates if curr_cands is None else curr_cands

        return [c for c in candidates if self.majority_prefers(c, cand)]

    def dominates(self, cand, curr_cands=None):
        """
        Returns the list of candidates that ``cand`` is majority preferred to in the majority graph restricted to ``curr_cands``.
        """
        candidates = self.candidates if curr_cands is None else curr_cands

        return [c for c in candidates if self.majority_prefers(cand, c)]

    def ratio(self, c1, c2):
        """
        Returns the ratio of the support of ``c1`` over ``c2`` to the support ``c2`` over ``c1``.
        """

        if self.support(c1, c2) > 0 and self.support(c2, c1) > 0:
            return self.support(c1, c2) / self.support(c2, c1)
        elif self.support(c1, c2) > 0 and self.support(c2, c1) == 0:
            return float(self.num_voters + self.support(c1, c2))
        elif self.support(c1, c2) == 0 and self.support(c2, c1) > 0:
            return 1 / (self.num_voters + self.support(c2, c1))
        elif self.support(c1, c2) == 0 and self.support(c2, c1) == 0:
            return 1

    def majority_prefers(self, c1, c2):
        """Returns True if ``c1`` is majority preferred to ``c2``."""

        return self.margin(c1, c2) > 0

    def strength_matrix(self, curr_cands = None, strength_function = None): 
        """
        Return the strength matrix of the profile.  The strength matrix is a matrix where the entry in row :math:`i` and column :math:`j` is the number of voters that rank the candidate with index :math:`i` over the candidate with index :math:`j`.  If ``curr_cands`` is provided, then the strength matrix is restricted to the candidates in ``curr_cands``.  If ``strength_function`` is provided, then the strength matrix is computed using the strength function."""
        
        if curr_cands is not None: 
            cindices = [cidx for cidx, _ in enumerate(curr_cands)]
            cindex_to_cand = lambda cidx: curr_cands[cidx]
            cand_to_cindex = lambda c: cindices[curr_cands.index(c)]
            strength_function = self.margin if strength_function is None else strength_function
            strength_matrix = np.array([[strength_function(cindex_to_cand(a_idx), cindex_to_cand(b_idx)) for b_idx in cindices] for a_idx in cindices])
        else:  
            cindices = self.cindices
            cindex_to_cand = self.cindex_to_cand
            cand_to_cindex = self.cand_to_cindex
            strength_matrix = np.array(self.margin_matrix) if strength_function is None else np.array([[strength_function(cindex_to_cand(a_idx), cindex_to_cand(b_idx)) for b_idx in cindices] for a_idx in cindices])

        return strength_matrix, cand_to_cindex

    def condorcet_winner(self, curr_cands=None):
        """Returns the Condorcet winner in the profile restricted to ``curr_cands`` if one exists, otherwise return None.

        The **Condorcet winner** is the candidate that is majority preferred to every other candidate.
        """
        curr_cands = curr_cands if curr_cands is not None else self.candidates

        cw = None
        for c in curr_cands:

            if all([self.majority_prefers(c, c1) for c1 in curr_cands if c1 != c]):
                cw = c
                break
        return cw

    def condorcet_loser(self, curr_cands=None):
        """Returns the Condorcet loser in the profile restricted to ``curr_cands`` if one exists, otherwise return None.

        A candidate :math:`c` is a  **Condorcet loser** if every other candidate  is majority preferred to :math:`c`.
        """

        curr_cands = curr_cands if curr_cands is not None else self.candidates

        cl = None
        for c1 in curr_cands:
            if all([self.majority_prefers(c2, c1) for c2 in curr_cands if c1 != c2]):
                cl = c1
                break  # if a Condorcet loser exists, then it is unique
        return cl

    def weak_condorcet_winner(self, curr_cands=None):
        """Returns a list of the weak Condorcet winners in the profile restricted to ``curr_cands`` (which may be empty).

        A candidate :math:`c` is a  **weak Condorcet winner** if there is no other candidate that is majority preferred to :math:`c`.

        .. note:: While the Condorcet winner is unique if it exists, there may be multiple weak Condorcet winners.
        """

        curr_cands = curr_cands if curr_cands is not None else self.candidates

        weak_cw = list()
        for c1 in curr_cands:
            if not any(
                [self.majority_prefers(c2, c1) for c2 in curr_cands if c1 != c2]
            ):
                weak_cw.append(c1)
        return sorted(weak_cw) if len(weak_cw) > 0 else None

    def copeland_scores(self, curr_cands = None, scores = (1,0,-1)):
        """The Copeland scores in the profile restricted to the candidates in ``curr_cands``. 

        The **Copeland score** for candidate :math:`c` is calculated as follows:  :math:`c` receives ``scores[0]`` points for every candidate that  :math:`c` is majority preferred to, ``scores[1]`` points for every candidate that is tied with :math:`c`, and ``scores[2]`` points for every candidate that is majority preferred to :math:`c`. The default ``scores`` is ``(1, 0, -1)``. 
        

        :param curr_cands: restrict attention to candidates in this list. Defaults to all candidates in the profile if not provided. 
        :type curr_cands: list[int], optional
        :param scores: the scores used to calculate the Copeland score of a candidate :math:`c`: ``scores[0]`` is for the candidates that :math:`c` is majority preferred to; ``scores[1]`` is the number of candidates tied with :math:`c`; and ``scores[2]`` is the number of candidate majority preferred to :math:`c`.  The default value is ``scores = (1, 0, -1)`` 
        :type scores: tuple[int], optional
        :returns: a dictionary associating each candidate in ``curr_cands`` with its Copeland score. 

        """        
    
        wscore, tscore, lscore = scores
        candidates = self.candidates if curr_cands is None else curr_cands
        c_scores = {c: 0.0 for c in candidates}
        for c1 in candidates:
            for c2 in candidates:
                if self.majority_prefers(c1, c2):
                    c_scores[c1] += wscore
                elif self.majority_prefers(c2, c1):
                    c_scores[c1] += lscore
                elif c1 != c2:
                    c_scores[c1] += tscore
        return c_scores


    def strict_maj_size(self):
        """Returns the strict majority of the number of voters."""

        return int(
            self.num_voters / 2 + 1
            if self.num_voters % 2 == 0
            else int(ceil(float(self.num_voters) / 2))
        )

    def plurality_scores(self, curr_cands=None):
        """
        Return the Plurality Scores of the candidates, assuming that each voter ranks a single candidate in first place.

        Parameters:
        - curr_cands: List of current candidates to consider. If None, use all candidates.

        Returns:
        - Dictionary with candidates as keys and their plurality scores as values.

        Raises:
        - ValueError: If any voter ranks multiple candidates in first place.
        """

        if curr_cands is None:
            curr_cands = self.candidates

        # Check if any voter ranks multiple candidates in first place
        if any(len(r.first(cs=curr_cands)) > 1 for r in self._rankings):
            raise ValueError("Cannot find the plurality scores unless all voters rank a unique candidate in first place.")

        rankings, rcounts = self.rankings_counts

        plurality_scores = {cand: 0 for cand in curr_cands}

        for ranking, count in zip(rankings, rcounts):
            first_place_candidates = ranking.first(cs=curr_cands)
            if len(first_place_candidates) == 1:
                cand = first_place_candidates[0]
                plurality_scores[cand] += count

        return plurality_scores

    def plurality_scores_ignoring_overvotes(self, curr_cands=None): 
        """
        Return the Plurality scores ignoring empty rankings and overvotes.
        """

        curr_cands = curr_cands if curr_cands is not None else self.candidates
        
        rankings, rcounts = self.rankings_counts
        
        return {cand: sum([c for r, c in zip(rankings, rcounts) if len(r.cands) > 0 and [cand] == r.first(cs=curr_cands)]) for cand in curr_cands}

    def borda_scores(self, 
                     curr_cands=None, 
                     borda_score_fnc=symmetric_borda_scores):
        
        curr_cands = self.candidates if curr_cands is None else curr_cands
        restricted_prof = self.remove_candidates([c for c in self.candidates if c not in curr_cands])
        return borda_score_fnc(restricted_prof)

    def tops_scores(
            self, 
            curr_cands=None, 
            score_type='approval'):
        """
        Return the tops scores of the candidates. 

        Parameters:
        - curr_cands: List of current candidates to consider. If None, use all candidates.
        - score_type: Type of tops score to compute. Options are 'approval' or 'split'.

        Returns:
        - Dictionary with candidates as keys and their tops scores as values.
        """

        if curr_cands is None:
            curr_cands = self.candidates

        rankings, rcounts = self.rankings_counts

        if score_type not in {'approval', 'split'}:
            raise ValueError("Invalid score_type specified. Use 'approval' or 'split'.")

        tops_scores = {cand: 0 for cand in curr_cands}

        if score_type == 'approval':
            for ranking, count in zip(rankings, rcounts):
                for cand in curr_cands:
                    if cand in ranking.first(cs=curr_cands):
                        tops_scores[cand] += count

        elif score_type == 'split':
            for ranking, count in zip(rankings, rcounts):
                for cand in curr_cands:
                    if cand in ranking.first(cs=curr_cands):
                        tops_scores[cand] += count * 1/len(ranking.first(cs=curr_cands))

        return tops_scores
        
    def remove_empty_rankings(self): 
        """
        Remove the empty rankings from the profile. 
        """
        new_rankings = list()
        new_rcounts = list()
                
        for r,c in zip(*(self.rankings_counts)):
            
            if len(r.cands) != 0: 
                new_rankings.append(r)
                new_rcounts.append(c)

        self._rankings = new_rankings
        self.rcounts = new_rcounts
        
        # update the number of voters
        self.num_voters = np.sum(self.rcounts)
        
        if self.using_extended_strict_preference: 
            self.use_extended_strict_preference()
        else: 
            self.use_strict_preference()

    def truncate_overvotes(self): 
        """Return a new profile in which all rankings with overvotes are truncated. """
        
        new_profile = copy.deepcopy(self)
        rankings, rcounts = new_profile.rankings_counts
        
        report = []
        for r,c in zip(rankings, rcounts): 
            old_ranking = copy.deepcopy(r)
            if r.has_overvote(): 
                r.truncate_overvote()
                report.append((old_ranking, r, c))
    
        if self.using_extended_strict_preference: 
            new_profile.use_extended_strict_preference()
        else: 
            new_profile.use_strict_preference()
            
        return new_profile, report

    def add_unranked_candidates(self): 
        """
        Return a profile in which for each voter, any unranked candidate is added to the bottom of their ranking. 
        """
        cands = self.candidates
        ranks = list()
        rcounts = list()

        for r in self._rankings: 
            min_rank = max(r.ranks) if len(r.ranks) > 0 else 1   
            new_r ={c:r for c, r in  r.rmap.items()}
            for c in cands: 
                if c not in new_r.keys(): 
                    new_r[c] = min_rank+1
            new_ranking = Ranking(new_r)

            found_it = False
            for _ridx, _r in enumerate(ranks):
                if new_ranking == _r: 
                    rcounts[_ridx] += 1
                    found_it = True
            if not found_it: 
                ranks.append(new_ranking)
                rcounts.append(1)

        return ProfileWithTies([r.rmap for r in ranks], rcounts=rcounts, cmap=self.cmap)

    @property
    def is_truncated_linear(self):
        """
        Return True if the profile only contains (truncated) linear orders.
        """
        return all([r.is_truncated_linear(len(self.candidates)) or r.is_linear(len(self.candidates)) for r in self._rankings])
    
    def to_linear_profile(self):
        """Return a linear profile from the profile with ties. If the profile is not a linear profile, then return None. 
        
        Note that the candidates in a Profile must be integers, so the candidates in the linear profile will be the indices of the candidates in the original profile.
        
        """
        rankings, rcounts = self.rankings_counts
        _new_rankings = [r.to_linear() for r in rankings]
        cand_to_cindx = {c:i for i,c in enumerate(sorted(self.candidates))}
        new_cmap = {cand_to_cindx[c]: self.cmap[c] for c in sorted(self.candidates)}
        if any([r is None or len(r) != len(self.candidates) for r in _new_rankings]): 
            print("Error: Cannot convert to linear profile.")
            return None
        new_rankings = [tuple([cand_to_cindx[c] for c in r]) for r in _new_rankings]
        return Profile(new_rankings, rcounts=rcounts, cmap=new_cmap)

    def replace_rankings(
            self, 
            old_ranking, 
            new_ranking, 
            num, 
            use_extended_strict_preference_for_comparison = False): 
        """

        Create a new profile by replacing num ballots matching old_ranking with new_ranking.

        If num is greater than the number of ballots matching old_ranking, then all ballots matching old_ranking are replaced with new_ranking.

        
        """
        using_extended_strict_pref = self.using_extended_strict_preference
        
        ranking_types, ranking_counts = self.rankings_counts

        if not isinstance(old_ranking, Ranking) or not isinstance(new_ranking, Ranking):
            raise ValueError("rankings must be of type Ranking")
            
        if use_extended_strict_preference_for_comparison:
            same_ranking = lambda r1, r2: same_ranking_extended_strict_pref(r1, r2, self.candidates)
        else:
            same_ranking = lambda r1, r2: r1 == r2
            
        new_ranking_types = []
        new_ranking_counts = []

        current_num = 0
        for r, c in zip(ranking_types, ranking_counts):
            
            if current_num < num and same_ranking(r, old_ranking):
                if c > num - current_num:
                    new_ranking_types.append(new_ranking)
                    new_ranking_counts.append(num - current_num)
                    new_ranking_types.append(old_ranking)
                    new_ranking_counts.append(c - (num - current_num))
                    current_num = num
                elif c == num - current_num and same_ranking(r, old_ranking):
                    new_ranking_types.append(new_ranking)
                    new_ranking_counts.append(num - current_num)
                    current_num = num
                elif c < num - current_num:
                    new_ranking_types.append(new_ranking)
                    new_ranking_counts.append(c)
                    current_num += c
            else:
                new_ranking_types.append(r)
                new_ranking_counts.append(c)

        new_prof = ProfileWithTies(new_ranking_types, new_ranking_counts, self.candidates, cmap=self.cmap)

        if using_extended_strict_pref:
            new_prof.use_extended_strict_preference()
    
        assert self.num_voters == new_prof.num_voters, "Problem: the number of voters is not the same in the new profile!"
        
        return new_prof   
    
    def margin_graph(self):
        """Returns the margin graph of the profile.  See :class:`.MarginGraph`.

        :Example:

        .. exec_code:: python

                from pref_voting.profiles_with_ties import ProfileWithTies
                prof = ProfileWithTies([{0: 1, 1: 2, 2: 3}, {1:1, 2:1, 0:2}, {2:1, 0:2}], [2, 3, 1])

                mg = prof.margin_graph()
                print(mg.edges)
                print(mg.margin_matrix)
        """

        return MarginGraph.from_profile(self)

    def support_graph(self):
        """Returns the support graph of the profile.  See :class:`.SupportGraph`.

        :Example:

        .. exec_code:: python

                from pref_voting.profiles_with_ties import ProfileWithTies
                prof = ProfileWithTies([{0: 1, 1: 2, 2: 3}, {1:1, 2:1, 0:2}, {2:1, 0:2}], [2, 3, 1])

                sg = prof.support_graph()
                print(sg.edges)
                print(sg.s_matrix)

        """

        return SupportGraph.from_profile(self)

    def majority_graph(self):
        """Returns the majority graph of the profile.  See :class:`.MarginGraph`.

        :Example:

        .. exec_code:: python

                from pref_voting.profiles_with_ties import ProfileWithTies
                prof = ProfileWithTies([{0: 1, 1: 2, 2: 3}, {1:1, 2:1, 0:2}, {2:1, 0:2}], [2, 3, 1])

                mg = prof.majority_graph()
                print(mg.edges)

        """

        return MajorityGraph.from_profile(self)

    def cycles(self):
        """Return a list of the cycles in the profile."""

        return self.margin_graph().cycles()

    def is_uniquely_weighted(self): 
        """Returns True if the profile is uniquely weighted. 
        
        A profile is **uniquely weighted** when there are no 0 margins and all the margins between any two candidates are unique.     
        """
        
        return MarginGraph.from_profile(self).is_uniquely_weighted()

    def remove_candidates(self, cands_to_ignore):
        """Remove all candidates from ``cands_to_ignore`` from the profile.

        :param cands_to_ignore: list of candidates to remove from the profile
        :type cands_to_ignore: list[int]
        :returns: a profile with candidates from ``cands_to_ignore`` removed.

        :Example:

        .. exec_code::

            from pref_voting.profiles_with_ties import ProfileWithTies
            prof = ProfileWithTies([{0: 1, 1: 2, 2: 3}, {1:1, 2:1, 0:2}, {2:1, 0:2}], [2, 3, 1])
            prof.display()
            new_prof = prof.remove_candidates([1])
            new_prof.display()
            print(new_prof.ranks)
        """

        updated_rankings = [
            {c: r for c, r in rank.rmap.items() if c not in cands_to_ignore}
            for rank in self._rankings
        ]
        new_candidates = [c for c in self.candidates if c not in cands_to_ignore]
        restricted_prof = ProfileWithTies(
            updated_rankings,
            rcounts=self.rcounts,
            candidates=new_candidates,
            cmap=self.cmap,
        )
        
        if self.using_extended_strict_preference: 
            restricted_prof.use_extended_strict_preference()
            
        return restricted_prof

    def report(self): 
        """
        Display a report of the types of rankings in the profile. 
        """
        num_ties = 0
        num_empty_rankings = 0
        num_with_skipped_ranks = 0
        num_trucated_linear_orders = 0
        num_linear_orders = 0
        
        rankings, rcounts = self.rankings_counts
        
        for r, c in zip(rankings, rcounts): 

            if r.has_tie():
                num_ties += c    
            if r.is_empty(): 
                num_empty_rankings += c
            elif r.is_linear(len(self.candidates)): 
                num_linear_orders += c
            elif r.is_truncated_linear(len(self.candidates)): 
                num_trucated_linear_orders += c
            
            if r.has_skipped_rank(): 
                num_with_skipped_ranks += c
        print(f'''There are {len(self.candidates)} candidates and {str(sum(rcounts))} {'ranking: ' if sum(rcounts) == 1 else 'rankings: '} 
        The number of empty rankings: {num_empty_rankings}
        The number of rankings with ties: {num_ties}
        The number of linear orders: {num_linear_orders}
        The number of truncated linear orders: {num_trucated_linear_orders}
        
The number of rankings with skipped ranks: {num_with_skipped_ranks}
        
        ''')

    def display_rankings(self): 
        """
        Display a list of the rankings in the profile. 
        """
        rankings, rcounts = self.rankings_counts
        
        rs = dict()
        for r, c in zip(rankings, rcounts): 
            if str(r) in rs.keys(): 
                rs[str(r)] += c
            else: 
                rs[str(r)] = c
                
        for r,c in rs.items(): 
            print(f"{r}: {c}")


    def anonymize(self): 
        """
        Return a profile which is the anonymized version of this profile. 
        """

        rankings = list()
        rcounts = list()
        for r in self.rankings:
            found_it = False
            for _ridx, _r in enumerate(rankings): 
                if r == _r: 
                    rcounts[_ridx] += 1
                    found_it = True
                    break
            if not found_it: 
                rankings.append(r)
                rcounts.append(1)
                
        prof = ProfileWithTies(rankings, rcounts=rcounts, cmap=self.cmap)

        if self.using_extended_strict_preference: 
            prof.use_extended_strict_preference()

        return prof

    def description(self): 
        """
        Return the Python code needed to create the profile.
        """
        return f"ProfileWithTies({[r.rmap for r in self._rankings]}, rcounts={[int(c) for c in self.rcounts]}, cmap={self.cmap})"

    def display(
            self, 
            cmap=None, 
            style="pretty", 
            curr_cands=None,
            order_by_counts=False,
            ):
        """Display a profile (restricted to ``curr_cands``) as an ascii table (using tabulate).

        :param cmap: the candidate map (overrides the cmap associated with this profile)
        :type cmap: dict[int,str], optional
        :param style: the candidate map to use (overrides the cmap associated with this profile)
        :type style: str ---  "pretty" or "fancy_grid" (or any other style option for tabulate)
        :param curr_cands: list of candidates
        :type curr_cands: list[int], optional
        :rtype: None

        :Example:

        .. exec_code::

            from pref_voting.profiles_with_ties import ProfileWithTies
            prof = ProfileWithTies([{0: 1, 1: 2, 2: 3}, {1:1, 2:1, 0:2}, {2:1, 0:2}], [2, 3, 1])
            prof.display()
            prof.display(cmap={0:"a", 1:"b", 2:"c"})

        """

        _rankings = copy.deepcopy(self._rankings)
        _rankings = [r.normalize_ranks() or r for r in _rankings]
        curr_cands = curr_cands if curr_cands is not None else self.candidates
        cmap = cmap if cmap is not None else self.cmap

        existing_ranks = list(range(min(min(r.ranks) for r in _rankings), max(max(r.ranks) for r in _rankings) + 1)) if len(_rankings) > 0 else []
        if order_by_counts: 
            _rankings, rcounts = zip(*sorted(zip(_rankings, self.rcounts), key=lambda x: x[1], reverse=True))
        else:
            rcounts = self.rcounts
        print(
            tabulate(
                [
                    [
                        " ".join(
                            [
                                str(cmap[c])
                                for c in r.cands_at_rank(rank)
                                if c in curr_cands
                            ]
                        )
                        for r in _rankings
                    ]
                    for rank in existing_ranks
                ],
                rcounts,
                tablefmt=style,
            )
        )

    def display_margin_graph(self, cmap=None, curr_cands=None):
        """
        Display the margin graph of the profile (restricted to ``curr_cands``) using the ``cmap``.  See :class:`.MarginGraph`.
        """

        cmap = cmap if cmap is not None else self.cmap
        MarginGraph.from_profile(self, cmap=cmap).display(curr_cands=curr_cands)

    def display_support_graph(self, cmap=None, curr_cands=None):
        """
        Display the support graph of the profile (restricted to ``curr_cands``) using the ``cmap``.  See :class:`.SupportGraph`.
        """

        cmap = cmap if cmap is not None else self.cmap
        SupportGraph.from_profile(self, cmap=cmap).display(curr_cands=curr_cands)

    def to_preflib_instance(self):
        """
        Returns an instance of the ``OrdinalInstance`` class from the ``preflibtools`` package. See ``pref_voting.io.writers.to_preflib_instance``.
        
        """
        from pref_voting.io.writers import to_preflib_instance

        return to_preflib_instance(self)

    @classmethod
    def from_preflib(
        cls, 
        instance_or_preflib_file, 
        include_cmap=False): 
        """
        Convert an preflib OrdinalInstance or file to a Profile.   See ``pref_voting.io.readers.from_preflib``.
        
        """
        from pref_voting.io.readers import preflib_to_profile

        return preflib_to_profile(
            instance_or_preflib_file, 
            include_cmap=include_cmap, 
            as_linear_profile=False)

    def write(
            self, 
            filename, 
            file_format="preflib", 
            csv_format="candidate_columns"):
        """
        Write a profile to a file.   See ``pref_voting.io.writers.write``.
        """
        from pref_voting.io.writers import  write
 
        return write(
            self, 
            filename, 
            file_format=file_format, 
            csv_format=csv_format)

    @classmethod
    def read(
        cls, 
        filename, 
        file_format="preflib",
        csv_format="candidate_columns",
        cand_type=None,
        items_to_skip=None): 
        """
        Read a profile from a file.  See ``pref_voting.io.readers.read``.
        
        """
        from pref_voting.io.readers import  read

        return read(
            filename, 
            file_format=file_format, 
            csv_format=csv_format,
            cand_type=cand_type,
            items_to_skip=items_to_skip,
            as_linear_profile=False,
            )
    
    def to_latex(self, cmap=None, curr_cands=None):
        """
        Returns a LaTeX table representation of the profile with ties.

        :param cmap: Dictionary mapping candidates to their names/labels. If None, use self.cmap.
        :param curr_cands: List of candidates to include in the table. If None, use all candidates.
        :return: A string containing the LaTeX table.
        """
        cmap = self.cmap if cmap is None else cmap
        curr_cands = self.candidates if curr_cands is None else curr_cands

        prof = copy.deepcopy(self)
        prof.remove_empty_rankings()

        _rankings = copy.deepcopy(prof._rankings)

        if len(_rankings) == 0: # if there are no rankings, return an empty string
            return ""
        
        _rankings = [r.normalize_ranks() or r for r in _rankings ]

        latex = "\\begin{tabular}{" + "c" * len(_rankings) + "}\n"
        latex += " & ".join(["$" + str(count) + "$" for count in self.rcounts]) + "\\\\\n"
        latex += "\\hline\n"

        max_rank = max(max(ranking.rmap.values()) for ranking in _rankings)

        for rank in range(1, max_rank + 1):
            row = []
            for ranking in _rankings:
                tied_cands = sorted([cmap[c] for c in curr_cands if c in ranking.rmap and ranking.rmap[c] == rank])
                if tied_cands:
                    row.append("$" + ",".join(tied_cands) + "$")
                else:
                    prev_cands = [c for c in curr_cands if c in ranking.rmap and ranking.rmap[c] < rank]
                    if prev_cands:
                        row.append(" ")
                    else:
                        row.append("$\\cdots$")
            latex += " & ".join(row) + "\\\\\n"

        latex += "\\end{tabular}"
        return latex

    def __eq__(self, other_prof): 
        """
        Returns true if two profiles are equal.  Two profiles are equal if they have the same rankings.  Note that we ignore the cmaps. 
        """

        rankings = self.rankings
        other_rankings = other_prof.rankings[:] # make a copy
        for r1 in rankings:
            for i, r2 in enumerate(other_rankings):
                if r1 == r2:   
                    # Remove the matched item to handle duplicates
                    del other_rankings[i]
                    break
            else:
                # If we didn't find a match for r1, the profiles are not identical
                return False
    
        return not other_rankings
    

    def __add__(self, other_prof): 
        """
        Returns the sum of two profiles.  The sum of two profiles is the profile that contains all the rankings from the first in addition to all the rankings from the second profile. 

        Note: the cmaps of the profiles are ignored. 
        """

        return ProfileWithTies(self._rankings + other_prof._rankings, rcounts=self.rcounts + other_prof.rcounts, candidates = sorted(list(set(self.candidates +other_prof.candidates))))



'''
    File: grade_methods.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: September 24, 2023
    
    Implementations of grading methods (also called evaluative methods). 
'''
from pref_voting.voting_method import  *
from itertools import product

@vm(name="Score Voting")
def score_voting(gprofile, curr_cands=None, evaluation_method="sum"):
    """Return the candidates with the largest scores, where scores are evaluated using the ``evaluation_method``, where the default is summing the scores of the candidates.  If ``curr_cands`` is provided, then the score vote is restricted to the candidates in ``curr_cands``.
    """
    
    curr_cands = gprofile.candidates if curr_cands is None else curr_cands
    if evaluation_method == "sum":
        evaluation_method_func = gprofile.sum
    elif evaluation_method == "mean" or evaluation_method == "average": 
        evaluation_method_func = gprofile.avg
    elif evaluation_method == "median": # returns lower median
        evaluation_method_func = gprofile.median

    scores = {
        c: evaluation_method_func(c) 
        for c in curr_cands if gprofile.has_grade(c)
        }

    max_score = max(scores.values())

    return sorted([c for c in scores.keys() if scores[c] == max_score])

@vm(name="Approval")
def approval(gprofile, curr_cands=None):
    """Return the approval vote of the grade profile ``gprofile``.  If ``curr_cands`` is provided, then the approval vote is restricted to the candidates in ``curr_cands``.

    .. warning:: 
        Approval Vote only works on Grade Profiles that are based on 2 grades: 0 and 1.

    """
    assert sorted(gprofile.grades) == [0, 1], "The  grades in the profile must be {0, 1}."
    
    return score_voting(gprofile, curr_cands=curr_cands, evaluation_method="sum")

@vm(name="Dis&approval")
def dis_and_approval(gprofile, curr_cands=None):
    """Return the Dis&approval vote of the grade profile ``gprofile``.  If ``curr_cands`` is provided, then the dis&approval vote is restricted to the candidates in ``curr_cands``.  See https://link.springer.com/article/10.1007/s00355-013-0766-7 for more information.

    .. warning:: 
        Dis&approval only works on Grade Profiles that are based on 2 grades: -1 and 1.

    """
    assert sorted(gprofile.grades) == [-1, 0, 1], "The  grades in the profile must be {-1, 0, 1}."
    
    return score_voting(gprofile, curr_cands=curr_cands, evaluation_method="sum")

@vm(name="Cumulative Voting")
def cumulative_voting(gprofile, curr_cands=None, max_total_grades=5):
    """Return the cumulative vote winner of the grade profile ``gprofile``.   This is the candidates with the largest sum of the grades where each voter submits a ballot of scores that sum to ``max_total_grades``.   If ``curr_cands`` is provided, then the cumulative vote is restricted to the candidates in ``curr_cands``."""
    assert sorted(gprofile.grades) == list(range(max_total_grades + 1)) and np.sum(gprofile.grades) == max_total_grades , f"For cumulative voting, the sum the grades must be {max_total_grades}."
    
    return score_voting(gprofile, curr_cands=curr_cands, evaluation_method="sum")

@vm(name="STAR")
def star(gprofile, curr_cands=None):
    """ Identify the top two candidates according to the sum of the grades for each candidate. Then hold a runoff between the top two candidates where the candidate that is ranked above the other by the most voters is the winner.  The candidates that move to the runoff round are: the candidate(s) with the largest sum of the grades and the candidate(s) with the 2nd largest sum of the grades (or perhaps tied for the largest sum of the grades). In the case of multiple candidates tied for the largest or 2nd largest sum of the grades, use parallel-universe tiebreaking: a candidate is a Star Vote winner if it is a winner in some head-to-head runoff as described. If the candidates are all tied for the  largest sum of the grades, then all candidates are winners. 

    See https://starvoting.us for more information.
    
    If ``curr_cands`` is provided, then the winners is restricted to the candidates in ``curr_cands``.

    .. warning:: 
        Star Vote only works on Grade Profiles that are based on 6 grades: 0, 1, 2, 3, 4, and 5.
    """

    assert sorted(gprofile.grades) == [0, 1, 2, 3, 4, 5], "The  grades in the profile must be {0, 1, 2, 3, 4, 5}."

    curr_cands = gprofile.candidates if curr_cands is None else curr_cands

    if len(curr_cands) == 1: 
        return list(curr_cands)

    cand_to_scores = {
        c: gprofile.sum(c) 
        for c in curr_cands if gprofile.has_grade(c)
        }
    
    scores = sorted(list(set(cand_to_scores.values())), reverse=True)

    max_score = scores[0]
    first = [c for c in cand_to_scores.keys() if cand_to_scores[c] == max_score]

    second = list()
    if len(first) == 1:
        second_score = scores[1]
        second = [c for c in cand_to_scores.keys() if cand_to_scores[c] == second_score]

    if len(second) > 0:
        all_runoff_pairs = product(first, second)
    else: 
        all_runoff_pairs = [(c1,c2) for c1,c2 in product(first, first) if c1 != c2]

    winners = list()
    for c1, c2 in all_runoff_pairs: 
        
        if gprofile.margin(c1,c2) > 0:
            winners.append(c1)
        elif gprofile.margin(c1,c2) < 0:
            winners.append(c2)
        elif gprofile.margin(c1,c2) == 0:
            winners.append(c1)
            winners.append(c2)
    
    return sorted(list(set(winners)))


def tiebreaker_diff(gprofile, cand, median_grade):
    """
    Tiebreaker when the there are multiple candidates with the largest median grade.  
    The tiebreaker is the difference between the proportion of voters who grade the candidate higher than the median grade and the proportion of voters who grade the candidate lower than the median grade.
    """
    prop_proponents = gprofile.proportion_with_higher_grade(cand, median_grade)
    prop_opponents = gprofile.proportion_with_lower_grade(cand, median_grade)

    return prop_proponents - prop_opponents

def tiebreaker_relative_shares(gprofile, cand, median_grade):
    """
    Tiebreaker when the there are multiple candidates with the largest median grade. 
    Returns the *relative shares* of the proponents and opponents of the candidate. 
    """

    prop_proponents = gprofile.proportion_with_higher_grade(cand, median_grade)
    prop_opponents = gprofile.proportion_with_lower_grade(cand, median_grade)

    return  (prop_proponents - prop_opponents) / (2 * (prop_proponents + prop_opponents))

def tiebreaker_normalized_difference(gprofile, cand, median_grade):
    """
    Tiebreaker when the there are multiple candidates with the largest median grade. 
    Returns the *normalized difference* of the proponents and opponents of the candidate. 
    """

    prop_proponents = gprofile.proportion_with_higher_grade(cand, median_grade)
    prop_opponents = gprofile.proportion_with_lower_grade(cand, median_grade)

    return  (prop_proponents - prop_opponents) /(2 * (1 - prop_proponents -  prop_opponents))


def tiebreaker_majority_judgement(gprofile, cand, median_grade):
    """
    Tiebreaker when the there are multiple candidates with the largest median grade. 
    Returns the proportion of voters assigning a higher grade than the median to cand if it is greater than the proportion of voters assigning a lower grade than the median to cand, otherwise return -1 * the proportion of voters assigning a lower grade than the median to cand. 
    """

    prop_proponents = gprofile.proportion_with_higher_grade(cand, median_grade)
    prop_opponents = gprofile.proportion_with_lower_grade(cand, median_grade)

    if prop_proponents > prop_opponents:
        return prop_proponents
    elif prop_opponents >= prop_proponents:
        return -prop_opponents


def greatest_median(gprofile, curr_cands=None, tb_func = tiebreaker_majority_judgement):

    """
    Returns the candidate(s) with the greatest median grade.  If there is a tie, the tie is broken by the tiebreaker function.
    
    """
    median_winners = score_voting(gprofile, curr_cands=curr_cands, evaluation_method="median")

    if len(median_winners) == 1:
        return median_winners
    else:
        tb_scores = {c: tb_func(gprofile, c, gprofile.median(c)) for c in median_winners}
        return sorted([c for c in tb_scores if tb_scores[c] == max(tb_scores.values())])

@vm(name="Majority Judgement")
def majority_judgement(gprofile, curr_cands=None):
    """
    The Majority Judgement voting method as describe in Balinski and Laraki (https://mitpress.mit.edu/9780262545716/majority-judgment/).
    """
    return greatest_median(gprofile, curr_cands=curr_cands, tb_func = tiebreaker_majority_judgement)


'''
    File: scoring_rules.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: January 6, 2022
    
    Implementations of scoring rules. 
'''
from pref_voting.voting_method import  *
from pref_voting.social_welfare_function import  *
from pref_voting.profiles import Profile
from pref_voting.rankings import Ranking, break_ties_alphabetically
from pref_voting.voting_method import _num_rank_last 
from pref_voting.profiles import _find_updated_profile, _num_rank
from pref_voting.weighted_majority_graphs import MarginGraph
from pref_voting.voting_method_properties import  ElectionTypes

@vm(name = "Plurality", 
    input_types=[ElectionTypes.PROFILE, 
                 ElectionTypes.TRUNCATED_LINEAR_PROFILE])
def plurality(profile, curr_cands = None):
    """The **Plurality score** of a candidate :math:`c` is the number of voters that rank :math:`c` in first place. The Plurality winners are the candidates with the largest Plurality score in the ``profile`` restricted to ``curr_cands``.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. seealso::

        The method :meth:`pref_voting.profiles.Profile.plurality_scores` returns a dictionary assigning the Plurality scores of each candidate. 
        
    :Example:

    .. exec_code::

        from pref_voting.profiles import Profile
        from pref_voting.scoring_methods import plurality
        
        prof1 = Profile([[0, 1, 2], [1, 0, 2], [2, 1, 0]], [3, 1, 2])
        prof1.display()
        print(plurality(prof1)) # [2]
        plurality.display(prof1)

        prof2 = Profile([[0, 1, 2], [1, 0, 2], [1, 2, 0]], [3, 1, 2])
        prof2.display()
        print(plurality(prof2)) # [0, 1]
        plurality.display(prof2)

    """
                                        
    curr_cands = profile.candidates if curr_cands is None else curr_cands

    # get the Plurality scores for all the candidates in curr_cands
    plurality_scores = profile.plurality_scores(curr_cands = curr_cands)
    
    assert plurality_scores != {}, "Cannot calculate plurality scores."

    max_plurality_score = max(plurality_scores.values())

    return sorted([c for c in curr_cands if plurality_scores[c] == max_plurality_score])


@swf(name="Plurality ranking")
def plurality_ranking(profile, curr_cands=None, local=True, tie_breaking=None):
    """The SWF that ranks the candidates in curr_cands according to their plurality scores. If local is True, then the plurality scores are computed with respect to the profile restricted to curr_cands. Otherwise, the plurality scores are computed with respect to the entire profile.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): The candidates to rank. If None, then all candidates in profile are ranked
        local (bool, optional): If True, then the plurality scores are computed with respect to the profile restricted to curr_cands. Otherwise, the plurality scores are computed with respect to the entire profile.
        tie_breaking (str, optional): The tie-breaking method to use. If None, then no tie-breaking is used. If "alphabetic", then the tie-breaking is done alphabetically.

    Returns:
        A Ranking object
    """

    cands = profile.candidates if curr_cands is None else curr_cands

    if local:
        plurality_scores_dict = profile.plurality_scores(curr_cands = cands)

    else:
        plurality_scores_dict = profile.plurality_scores()
        plurality_scores_dict = {k: v for k, v in plurality_scores_dict.items() if k in cands}

    assert plurality_scores_dict != {}, "Cannot calculate plurality scores."

    for cand in cands:
        plurality_scores_dict[cand] = -plurality_scores_dict[cand]

    p_ranking = Ranking(plurality_scores_dict)
    p_ranking.normalize_ranks()

    if tie_breaking == "alphabetic":
        p_ranking = break_ties_alphabetically(p_ranking)

    return p_ranking

@vm(name = "Borda",
    input_types=[ElectionTypes.PROFILE, ElectionTypes.MARGIN_GRAPH])
def borda(edata, curr_cands = None, algorithm = "positional"):
    """The **Borda score** of a candidate is calculated as follows: If there are :math:`m` candidates, then the Borda score of candidate :math:`c` is :math:`\sum_{r=1}^{m} (m - r) * Rank(c,r)` where :math:`Rank(c,r)` is the number of voters that rank candidate :math:`c` in position :math:`r`. The Borda winners are the candidates with the largest Borda score in the ``profile`` restricted to ``curr_cands``. 
    Args:
        edata (Profile, MarginGraph): An anonymous profile of linear orders or a MarginGraph.
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``
        algorithm (String): if "positional", then the Borda score of a candidate is calculated from each voter's ranking as described above. If "marginal", then the Borda score of a candidate is calculated as the sum of the margins of the candidate vs. all other candidates. The positional scores and marginal scores are affinely equivalent.

    Returns: 
        A sorted list of candidates

    .. note:
        If edata is a MarginGraph, then the "marginal" algorithm is used.

    .. seealso::

        The method :meth:`pref_voting.profiles.Profile.borda_scores` returns a dictionary assigning the Borda score to each candidate. 
        
    :Example:

    .. exec_code::

        from pref_voting.profiles import Profile
        from pref_voting.scoring_methods import borda
        
        prof1 = Profile([[0, 1, 2], [1, 0, 2], [2, 1, 0]], [3, 1, 2])
        prof1.display()
        print(borda(prof1)) # [0,1]
        borda.display(prof1)

        prof2 = Profile([[0, 1, 2], [1, 0, 2], [1, 2, 0]], [3, 1, 2])
        prof2.display()
        print(borda(prof2)) # [1]
        borda.display(prof2)

    """

    curr_cands = edata.candidates if curr_cands is None else curr_cands

    if isinstance(edata,MarginGraph):
        algorithm = "marginal"

    if algorithm == "positional":
        # get the Borda scores for all the candidates in curr_cands
        borda_scores = edata.borda_scores(curr_cands = curr_cands)

    if algorithm == "marginal":
        borda_scores = {c: sum([edata.margin(c,d) for d in curr_cands]) for c in curr_cands}
    
    max_borda_score = max(borda_scores.values())
    
    return sorted([c for c in curr_cands if borda_scores[c] == max_borda_score])

@swf(name="Borda ranking")
def borda_ranking(profile, curr_cands=None, local=True, tie_breaking=None):
    """The SWF that ranks the candidates in curr_cands according to their Borda scores. If local is True, then the Borda scores are computed with respect to the profile restricted to curr_cands. Otherwise, the Borda scores are computed with respect to the entire profile.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): The candidates to rank. If None, then all candidates in profile are ranked
        local (bool, optional): If True, then the Borda scores are computed with respect to the profile restricted to curr_cands. Otherwise, the Borda scores are computed with respect to the entire profile.
        tie_breaking (str, optional): The tie-breaking method to use. If None, then no tie-breaking is used. If "alphabetic", then the tie-breaking is done alphabetically.

    Returns:
        A Ranking object
    """

    cands = profile.candidates if curr_cands is None else curr_cands

    if local:
        borda_scores_dict = profile.borda_scores(curr_cands = cands)

    else:
        borda_scores_dict = profile.borda_scores()
        borda_scores_dict = {k: v for k, v in borda_scores_dict.items() if k in cands}

    for cand in cands:
        borda_scores_dict[cand] = -borda_scores_dict[cand]

    b_ranking = Ranking(borda_scores_dict)
    b_ranking.normalize_ranks()

    if tie_breaking == "alphabetic":
        b_ranking = break_ties_alphabetically(b_ranking)

    return b_ranking

@vm(name = "Anti-Plurality",
    input_types=[ElectionTypes.PROFILE])
def anti_plurality(profile, curr_cands = None):
    """The **Anti-Plurality score** of a candidate $c$ is the number of voters that rank $c$ in last place.  The Anti-Plurality winners are the candidates with the smallest Anti-Plurality score in the ``profile`` restricted to ``curr_cands``. 

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    :Example:

    .. exec_code::

        from pref_voting.profiles import Profile
        from pref_voting.scoring_methods import anti_plurality
        
        prof1 = Profile([[2, 1, 0], [2, 0, 1], [0, 1, 2]], [3, 1, 2])
        prof1.display()
        print(anti_plurality(prof1)) # [1]
        anti_plurality.display(prof1)

        prof2 = Profile([[2, 1, 0], [2, 0, 1], [0, 2, 1]], [3, 1, 2])
        prof2.display()
        print(anti_plurality(prof2)) # [2]
        anti_plurality.display(prof2)

    """
    
    # get ranking data
    rankings, rcounts = profile.rankings_counts

    curr_cands = profile.candidates if curr_cands is None else curr_cands
    cands_to_ignore = np.array([c for c in profile.candidates if c not in curr_cands])
    
    last_place_scores = {c: _num_rank_last(rankings, rcounts, cands_to_ignore, c) for c in curr_cands}
    min_last_place_score = min(list(last_place_scores.values()))
    
    return sorted([c for c in curr_cands if last_place_scores[c] == min_last_place_score])

@swf(name="Anti-Plurality ranking")
def anti_plurality_ranking(profile, curr_cands=None, local=True, tie_breaking=None):
    """The SWF that ranks the candidates in curr_cands according to their Anti-Plurality scores. If local is True, then the Anti-Plurality scores are computed with respect to the profile restricted to curr_cands. Otherwise, the Anti-Plurality scores are computed with respect to the entire profile.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): The candidates to rank. If None, then all candidates in profile are ranked
        local (bool, optional): If True, then the Anti-Plurality scores are computed with respect to the profile restricted to curr_cands. Otherwise, the Anti-Plurality scores are computed with respect to the entire profile.
        tie_breaking (str, optional): The tie-breaking method to use. If None, then no tie-breaking is used. If "alphabetic", then the tie-breaking is done alphabetically.

    Returns:
        A Ranking object
    """

    cands = profile.candidates if curr_cands is None else curr_cands

    rankings, rcounts = profile.rankings_counts

    if local:
        cands_to_ignore = np.array([c for c in profile.candidates if c not in cands])

    else:
        cands_to_ignore = np.array([])
    
    anti_plurality_scores_dict = {c: _num_rank_last(rankings, rcounts, cands_to_ignore, c) for c in cands}

    ap_ranking = Ranking(anti_plurality_scores_dict)
    ap_ranking.normalize_ranks()

    if tie_breaking == "alphabetic":
        ap_ranking = break_ties_alphabetically(ap_ranking)

    return ap_ranking


@vm(name = "Scoring Rule",
    skip_registration=True,)
def scoring_rule(profile, curr_cands = None, score = lambda num_cands, rank : 1 if rank == 1 else 0):
    """A general scoring rule.  Each voter assign a score to each candidate using the ``score`` function based on their submitted ranking (restricted to candidates in ``curr_cands``).   Returns that candidates with the greatest overall score in the profile restricted to ``curr_cands``. 

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``
        score (function): A function that accepts two parameters ``num_cands`` (the number of candidates) and ``rank`` (a rank of a candidate) used to calculate the score of a candidate.   The default ``score`` function assigns 1 to a candidate ranked in first place, otherwise it assigns 0 to the candidate. 

    Returns: 
        A sorted list of candidates

    .. important:: 
        The signature of the ``score`` function is:: 

            def score(num_cands, rank):
                # return an int or float 

    :Example:

    .. exec_code::

        from pref_voting.profiles import Profile
        from pref_voting.scoring_methods import scoring_rule, plurality, borda, anti_plurality
        
        prof = Profile([[0, 1, 2], [1, 0, 2], [2, 1, 0]], [3, 1, 2])
        prof.display()
        scoring_rule.display(prof) # Uses default scoring function, same a Plurality        
        plurality.display(prof)

        scoring_rule.display(prof, score=lambda num_cands, rank: num_cands - rank) # same a Borda
        borda.display(prof)

        scoring_rule.display(prof, score=lambda num_cands, rank: -1 if rank == num_cands else 0) # same as Anti-Plurality
        anti_plurality.display(prof)

    """
    
    # get ranking data
    _rankings, rcounts = profile.rankings_counts

    # get (restricted) rankings
    cands_to_ignore = np.array([c for c in profile.candidates if c not in curr_cands]) if curr_cands is not None else np.array([])
    rankings = _rankings if curr_cands is None else _find_updated_profile(np.array(_rankings), cands_to_ignore, len(profile.candidates))
    candidates = profile.candidates if curr_cands is None else curr_cands

    # find the candidate scores using the score function
    cand_scores = {c: sum(_num_rank(rankings, rcounts, c, level) * score(len(candidates), level) for level in range(1, len(candidates) + 1)) for c in candidates}
    
    # find maximum score
    max_score = max(cand_scores.values())

    return sorted([c for c in candidates if cand_scores[c] == max_score])

def create_scoring_method(score, name):
    """Create a scoring method using a given score function and name."""

    def _vm(profile, curr_cands = None):
        return scoring_rule(profile, curr_cands = curr_cands, score = score)

    return VotingMethod(_vm, name = name)

@vm(name = "Dowdall",
    input_types=[ElectionTypes.PROFILE])
def dowdall(profile, curr_cands = None):
    """The first-ranked candidate gets 1 point, the second-ranked candidate gets 1/2 point, the third-ranked candidate gets 1/3 point, and so on.  The Dowdall winners are the candidates with the greatest overall score in the profile restricted to ``curr_cands``.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``.

    Returns:
        A sorted list of candidates

    .. note::
        This system is used in Nauru. See, e.g., Jon Fraenkel & Bernard Grofman (2014), "The Borda Count and its real-world alternatives: Comparing scoring rules in Nauru and Slovenia," Australian Journal of Political Science, 49:2, 186-205, DOI: 10.1080/10361146.2014.900530.
    
    """

    return scoring_rule(profile, curr_cands = curr_cands, score = lambda num_cands, rank: 1 / rank)

@vm(name="Positive-Negative Voting",
    input_types=[ElectionTypes.PROFILE])
def positive_negative_voting(profile, curr_cands = None):
    """The **Positive-Negative Voting** method is a scoring rule where each voter assigns a score of 1 to their top-ranked candidate and a score of -1 to their bottom-ranked candidate.  See https://onlinelibrary.wiley.com/doi/10.1111/ecin.12929 for more information.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns:
        A sorted list of candidates

    """

    return scoring_rule(profile, 
                        curr_cands = curr_cands, 
                        score = lambda num_cands, rank : 1 if rank == 1 else (-1 if rank == num_cands else 0))


@swf(name = "Score Ranking")
def score_ranking(profile, curr_cands = None, score = lambda num_cands, rank : 1 if rank == 1 else 0):
    """A general swf that ranks the candidates according to the given score function.  Each voter assign a score to each candidate using the ``score`` function based on their submitted ranking (restricted to candidates in ``curr_cands``).   Returns the ranking of the candidates according to the scores. 

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``
        score (function): A function that accepts two parameters ``num_cands`` (the number of candidates) and ``rank`` (a rank of a candidate) used to calculate the score of a candidate.   The default ``score`` function assigns 1 to a candidate ranked in first place, otherwise it assigns 0 to the candidate. 

    Returns: 
        A ranking of the candidates

    .. important:: 
        The signature of the ``score`` function is:: 

            def score(num_cands, rank):
                # return an int or float 
    """
    
    # get ranking data
    _rankings, rcounts = profile.rankings_counts

    # get (restricted) rankings
    cands_to_ignore = np.array([c for c in profile.candidates if c not in curr_cands]) if curr_cands is not None else np.array([])
    rankings = _rankings if curr_cands is None else _find_updated_profile(np.array(_rankings), cands_to_ignore, len(profile.candidates))
    candidates = profile.candidates if curr_cands is None else curr_cands

    # find the candidate scores using the score function
    cand_scores = {c: -1 * sum(_num_rank(rankings, rcounts, c, level) * score(len(candidates), level) for level in range(1, len(candidates) + 1)) for c in candidates}
    
    ranking = Ranking(cand_scores)
    ranking.normalize_ranks()
    return ranking

## Borda for ProfilesWithTies

def symmetric_borda_scores(profile): 
    """
    The symmetric Borda score of a candidate c for a ranking r is the number of candidates ranked strictly below c according to r
    minus the number of candidates ranked strictly above c according to r. 
    
    See http://www.illc.uva.nl/~ulle/pubs/files/TerzopoulouEndrissJME2021.pdf for a discussion. 
    """
    
    return  {cand: sum([len([_cand for _cand in profile.candidates if r.extended_strict_pref(cand, _cand)]) * c 
                    for r,c in zip(*profile.rankings_counts)]) - sum([len([_cand for _cand in profile.candidates if r.extended_strict_pref(_cand, cand)]) * c 
                    for r,c in zip(*profile.rankings_counts)]) for cand in profile.candidates}

def domination_borda_scores(profile): 
    """
    The domination Borda score of a candidate c for a ranking r is the number of candidates ranked strictly below c according to r. 
    
    See http://www.illc.uva.nl/~ulle/pubs/files/TerzopoulouEndrissJME2021.pdf for a discussion. 

    """
    
    return  {cand: sum([len([_cand for _cand in profile.candidates if r.extended_strict_pref(cand, _cand)]) * c 
                    for r,c in zip(*profile.rankings_counts)]) for cand in profile.candidates}


def weak_domination_borda_scores(profile): 
    """
    The weak domination Borda score of a candidate c for a ranking r is the number of candidates ranked weakly below c according to r. 
    
    See http://www.illc.uva.nl/~ulle/pubs/files/TerzopoulouEndrissJME2021.pdf for a discussion. 

    """
    
    return  {cand: sum([len([_cand for _cand in profile.candidates if r.extended_weak_pref(cand, _cand)]) * c 
                    for r,c in zip(*profile.rankings_counts)]) for cand in profile.candidates}

def non_domination_borda_scores(profile): 
    """
    The non-domination Borda score of a candidate c for a ranking r is -1 times the number of candidates ranked strictly above c according to r. 
    
    See http://www.illc.uva.nl/~ulle/pubs/files/TerzopoulouEndrissJME2021.pdf for a discussion. 

    """
    
    return  {cand: -sum([len([_cand for _cand in profile.candidates if r.extended_strict_pref(_cand, cand)]) * c 
                    for r,c in zip(*profile.rankings_counts)]) for cand in profile.candidates}


@vm(name="Borda (for Truncated Profiles)",
    input_types=[ElectionTypes.TRUNCATED_LINEAR_PROFILE])
def borda_for_profile_with_ties(
    profile, 
    curr_cands=None, 
    borda_scores=symmetric_borda_scores): 
    """
    Borda score for truncated linear orders using different ways of defining the Borda score for truncated linear
    orders.  
    """
    # profile must be a ProfileWithTies object
    if isinstance(profile, Profile): 
        return borda(profile, curr_cands = curr_cands)

    curr_cands = curr_cands if curr_cands is not None else profile.candidates 
    
    restricted_prof = profile.remove_candidates([c for c in profile.candidates if c not in curr_cands])
    
    b_scores = borda_scores(restricted_prof)
    
    max_borda_score = max(b_scores.values())
    
    return sorted([c for c in restricted_prof.candidates if b_scores[c] == max_borda_score])


scoring_swfs = [
    plurality_ranking,
    borda_ranking,
    anti_plurality_ranking,
    score_ranking   
]




from pref_voting.scoring_methods import *
from pref_voting.iterative_methods import *
from pref_voting.c1_methods import *
from pref_voting.margin_based_methods import *
from pref_voting.combined_methods import *
from pref_voting.other_methods import *
from pref_voting.stochastic_methods import *




'''
    File: voting_methods.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: November 6, 2021
    Update: January 15, 2023
    
    The VotingMethodProperties class to encapsulate properties of voting methods.
'''
from enum import Enum

class ElectionTypes(Enum): 
    PROFILE = "Profile"
    PROFILE_WITH_TIES = "ProfileWithTies"
    TRUNCATED_LINEAR_PROFILE = "TruncatedLinearProfile"
    MAJORITY_GRAPH = "MajorityGraph"
    MARGIN_GRAPH = "MarginGraph"

class VotingMethodProperties:
    """
    Class to represent the properties of a voting method.
    """
    def __init__(self, **kwargs):
        """
        Initialize the properties of the voting method.

        Args:
            **kwargs: Arbitrary keyword arguments representing properties and their boolean values.
        """
        for key, value in kwargs.items():
            setattr(self, key, value)

    def items(self):
        """
        Return all properties as a list of items.

        Returns:
            List of all properties and their boolean values.
        """
        return self.__dict__.items()

    def satisfied(self):
        """
        List all properties that are satisfied (i.e., True).

        Returns:
            List of property names that are satisfied.
        """
        return [key for key, value in self.items() if value is True]
    
    def violated(self):
        """
        List all properties that are violated (i.e., False).

        Returns:
            List of property names that are violated.
        """
        return [key for key, value in self.items() if value is False]

    def not_available(self):
        """
        List all properties that are not available (i.e., None).

        Returns:
            List of property names that are None (currently not available).
        """
        return [key for key, value in self.items() if value is None]

    
    def __getitem__(self, prop):
        """
        Get the value of a property.

        Args:
            prop: The name of the property.

        Returns:
            The value of the property, or None if the property does not exist.
        """
        return self.__dict__.get(prop, None)

    def __str__(self) -> str:
        """
        Return a string representation of all properties.

        Returns:
            A string representation of all properties with their status (Satisfied, Violated, or N/A).
        """
        properties = [
            f"{key}: {'Satisfied' if value is True else 'Violated' if value is False else 'N/A'}"
            for key, value in self.__dict__.items()
        ]
        return "\n".join(properties)
